from __future__ import absolute_import, unicode_literals

import os
import warnings
import logging

from django.conf import settings
from django.test.signals import setting_changed

from ppmutils import environment as env

# Always import this module as follows:
# from ppm_client import settings [as ppm_settings]

# Don't import directly CONFIG or PANELs, or you will miss changes performed
# with override_settings in tests.

# Set service URLs based on environment
PPM_ENVIRONMENTS = {
    "prod": {
        "API_URL": "https://api.ppm.aws.dbmi.hms.harvard.edu",
        "ADMIN_URL": "https://p2m2a.dbmi.hms.harvard.edu",
        "QUESTIONNAIRE_URL": "https://questionnaire.ppm.dbmi.hms.harvard.edu/questionnaire",
        "DATA_URL": "https://data.ppm.dbmi.hms.harvard.edu",
        "FHIR_URL": "https://fhir.ppm.aws.dbmi.hms.harvard.edu",
        "FILESERVICE_URL": "https://files.ppm.aws.dbmi.hms.harvard.edu",
        "DASHBOARD_URL": {
            "neer": "https://neer.ppm.aws.dbmi.hms.harvard.edu",
            "autism": "https://autism.ppm.aws.dbmi.hms.harvard.edu",
            "example": "https://example.ppm.aws.dbmi.hms.harvard.edu",
            "rant": "https://rant.ppm.aws.dbmi.hms.harvard.edu",
        },
    },
    "dev": {
        "API_URL": "https://api.ppm.aws.dbmi-dev.hms.harvard.edu",
        "ADMIN_URL": "https://p2m2a.aws.dbmi-dev.hms.harvard.edu",
        "QUESTIONNAIRE_URL": "https://questionnaire.ppm.aws.dbmi-dev.hms.harvard.edu/questionnaire",
        "DATA_URL": "https://data.ppm.aws.dbmi-dev.hms.harvard.edu",
        "FHIR_URL": "https://fhir.ppm.aws.dbmi-dev.hms.harvard.edu",
        "FILESERVICE_URL": "https://files.ppm.aws.dbmi-dev.hms.harvard.edu",
        "DASHBOARD_URL": {
            "neer": "https://neer.ppm.aws.dbmi-dev.hms.harvard.edu",
            "autism": "https://autism.ppm.aws.dbmi-dev.hms.harvard.edu",
            "example": "https://example.ppm.aws.dbmi-dev.hms.harvard.edu",
            "rant": "https://rant.ppm.aws.dbmi-dev.hms.harvard.edu",
        },
    },
    "local": {
        "API_URL": "https://api.ppm.aws.dbmi-loc.hms.harvard.edu",
        "ADMIN_URL": "https://p2m2a.aws.dbmi-loc.hms.harvard.edu",
        "QUESTIONNAIRE_URL": "https://questionnaire.ppm.aws.dbmi-loc.hms.harvard.edu/questionnaire",
        "DATA_URL": "https://data.ppm.aws.dbmi-loc.hms.harvard.edu",
        "FHIR_URL": "https://fhir.ppm.aws.dbmi-loc.hms.harvard.edu",
        "FILESERVICE_URL": "https://files.ppm.aws.dbmi-loc.hms.harvard.edu",
        "DASHBOARD_URL": {
            "neer": "https://neer.ppm.aws.dbmi-loc.hms.harvard.edu",
            "autism": "https://autism.ppm.aws.dbmi-loc.hms.harvard.edu",
            "example": "https://example.ppm.aws.dbmi-loc.hms.harvard.edu",
            "rant": "https://rant.ppm.aws.dbmi-loc.hms.harvard.edu",
        },
    },
}

CONFIG_DEFAULTS = {
    # The identifier for this service and/or project
    "STUDY": None,
    # Client options, assume production environment
    "ENVIRONMENT": "prod",
    # Set prod URLs
    "API_URL": None,
    "ADMIN_URL": None,
    "QUESTIONNAIRE_URL": None,
    "DATA_URL": None,
    "FHIR_URL": None,
    "FILESERVICE_URL": None,
    # Optionally disable logging
    "LOGGER_NAME": "ppmutils",
    "ENABLE_LOGGING": True,
    "LOG_LEVEL": logging.WARNING,
    # Configure email settings
    "EMAIL_ENABLED": False,
    "EMAIL_DEFAULT_FROM": "ppm-no-reply@dbmi.hms.harvard.edu",
    "EMAIL_TEST_ACCOUNTS": [r"tester[\d]?@peoplepoweredmedicine.org:admin@peoplepoweredmedicine.org"],
    "EMAIL_SIGNATURE": {
        "name": "Jane Smith",
        "email": "jane_smith@peoplepoweredmedicine.org",
        "title": "Manager, People-Powered Medicine",
        "phone": "617-555-5555",
    },
}

# List of settings that cannot be defaulted and must be user-defined
REQUIRED_SETTINGS = ("FHIR_URL",)

# List of settings that have been removed
REMOVED_SETTINGS = ()

# List of service URLs
SERVICE_URLS = (
    "API_URL",
    "FHIR_URL",
    "ADMIN_URL",
    "QUESTIONNAIRE_URL",
    "DATA_URL",
    "FILESERVICE_URL",
)

# List of needed email settings if enabled
EMAIL_SETTINGS = (
    "EMAIL_BACKEND",
    "EMAIL_HOST",
    "EMAIL_HOST_USER",
    "EMAIL_HOST_PASSWORD",
    "EMAIL_PORT",
    "EMAIL_USE_SSL",
)


class PPMSettings(object):
    """
    A settings object, that allows PPMUtils Client settings to be accessed as properties.
    For example:
        from ppmutils.settings import ppm_settings
        print(ppm_settings.STUDY)
    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, defaults=None):
        self.defaults = defaults or CONFIG_DEFAULTS
        self._cached_attrs = set()

    @property
    def user_settings(self):

        # Check to see if user configs have been loaded or not
        if not hasattr(self, "_user_settings"):

            try:
                # Load user-specified configurations
                user_settings = getattr(settings, "PPM_CONFIG", {})

                # Update the client config with pre-defined environment URLs, etc
                if user_settings.get("ENVIRONMENT") in PPM_ENVIRONMENTS:
                    env_settings = PPM_ENVIRONMENTS[user_settings.get("ENVIRONMENT")]
                    user_settings.update({k: v for k, v in env_settings.items() if k not in user_settings})

                else:
                    # Check for them in environment
                    for key in PPM_ENVIRONMENTS["prod"].keys():
                        if os.environ.get("PPM_{}".format(key.upper())):
                            user_settings[key] = env.get_str("PPM_{}".format(key.upper()))

                # Check them
                self._user_settings = self.__check_user_settings(user_settings)

            except Exception as e:
                raise SystemError("PPMUtils settings are invalid: {}".format(e))

        return self._user_settings

    def dashboard_url(self, study):
        """
        Returns the defined dashboard for the passed study
        :param study: The study we want the dashboard URL for
        :type study: str
        :return: The dashboard URL
        :rtype: str
        """
        try:
            # Get environments
            return PPM_ENVIRONMENTS[self.ENVIRONMENT]["DASHBOARD_URL"][study]

        except (ValueError, IndexError):
            warnings.warn(f'No defined PPM dashboard URL for "{self.ENVIRONMENT}/{study}"')

        return None

    def __getattr__(self, attr):

        # Any attribute must be in either required settings or defaults
        if attr not in REQUIRED_SETTINGS and attr not in self.defaults:
            raise AttributeError("Invalid PPM setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:

            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)

        return val

    def __check_user_settings(self, user_settings):
        SETTINGS_DOC = "https://github.com/hms-dbmi/ppmutils"

        for setting in REMOVED_SETTINGS:
            if setting in user_settings:
                raise RuntimeError(
                    "The '%s' setting has been removed. Please refer to '%s' for available settings."
                    % (setting, SETTINGS_DOC)
                )

        # Ensure environment is set and if not prod or dev, ensure service URLs are provided
        if (
            "ENVIRONMENT" in user_settings
            and user_settings["ENVIRONMENT"].lower() != "prod"
            and user_settings["ENVIRONMENT"].lower() != "dev"
        ):
            warnings.warn("ENVIRONMENT is not set to production, this should be a test environment", ResourceWarning)

            # Check URLs
            self._url_settings(user_settings)

        # Check email settings
        self._email_settings(user_settings)

        return user_settings

    def _url_settings(self, user_settings):
        """
        Checks the host settings for properly configured URLs and adds the correct
        configurations if needed.
        """
        # Iterate service URLs and ensure they're defined, error out on required ones
        missing_urls = []
        for url in SERVICE_URLS:

            # Check if not added
            if url not in user_settings:

                # Check environment
                if os.environ.get(f"PPM_{url}"):
                    user_settings[url] = os.environ.get(f"PPM_{url}")
                elif url in REQUIRED_SETTINGS:
                    missing_urls.append(url)
                else:
                    warnings.warn(f'PPM URLs: Env "{url}"/"PPM_{url}" is missing but not required')

        # Error out if needed
        if missing_urls:
            raise AttributeError("{} configuration(s) must be set".format(missing_urls))

    def _email_settings(self, user_settings):
        """
        Checks the host settings for properly configured email backends and adds the correct
        configurations if needed.
        """
        if user_settings.get("EMAIL_ENABLED"):

            # Check and remap remaining settings
            missing_settings = []
            for email_setting in EMAIL_SETTINGS:

                # Ensure settings has these defined
                if getattr(settings, email_setting) is None:
                    missing_settings.append(email_setting)

            if missing_settings:
                raise AttributeError("PPM Email: Required email settings are missing: {}".format(missing_settings))

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")

    def __dir__(self):

        # Return defaults and user configs patched over
        attrs = dict()
        attrs.update(self.defaults)
        attrs.update(self.user_settings)

        return attrs.keys()

    def get_logger(self):
        """
        Returns the logger and manages whether logs propogate or not depending on user configs
        :return: logger
        """
        logger = logging.getLogger(__name__)

        # Check if disabled
        if not self.ENABLE_LOGGING:
            logger.propagate = False

        return logger


# Create the instance by which the settings should be accessed
ppm_settings = PPMSettings(CONFIG_DEFAULTS)

# Configure logging
logger = logging.getLogger(ppm_settings.LOGGER_NAME)
logger.disabled = not ppm_settings.ENABLE_LOGGING
logger.setLevel(ppm_settings.LOG_LEVEL)


def reload_ppm_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "PPM_CONFIG":
        ppm_settings.reload()


setting_changed.connect(reload_ppm_settings)



"""
PPM Utils
"""

__title__ = "PPM Utils"
__version__ = "1.0.0-beta.6"
__author__ = "Bryan Larson"
__license__ = "BSD 2-Clause"
__copyright__ = "Copyright 2011-2020 Harvard Medical School Department of Biomedical Informatics"

# Version synonym
VERSION = __version__

# Header encoding (see RFC5987)
HTTP_HEADER_ENCODING = "iso-8859-1"

# Default datetime input and output formats
ISO_8601 = "iso-8601"

# Set the Django app
default_app_config = "ppmutils.apps.PPMUtilsConfig"



from ppmutils.ppm import PPM
from ppmutils.fhir import FHIR

import logging

logger = logging.getLogger(__name__)


class Auth(object):

    ITEM = "ppm"
    ADMIN = "admin"
    VIEW = "view"
    PERMISSIONS = [ADMIN, VIEW]
    METHOD_PERMISSIONS = {
        "HEAD": [ADMIN, VIEW],
        "OPTION": [ADMIN, VIEW],
        "GET": [ADMIN, VIEW],
        "POST": [ADMIN],
        "PATCH": [ADMIN],
        "PUT": [ADMIN],
        "DELETE": [ADMIN],
    }

    @classmethod
    def get_permission(cls, permissions, method=None, study=None):
        """
        Inspects the set of permissions and returns the permission
        valid for the request. If a study is passed, this method
        returns True if permissions are on the specific study or PPM.
        If method is passed, the permissioned is checked for that
        operation.

        :param permissions: A list of permissions from DBMI-AuthZ
        :type permissions: list
        :param method: The requested method to check permissions for
        :type method: str
        :param study: A specific study, defaults to None
        :type study: str, optional
        :returns: Whether the request has permissions or not
        :rtype: bool
        """

        # Map permissions
        map = {a["item"].lower(): a["permission"].lower() for a in permissions}

        # Set placeholders for item and permission. We want to return the
        # upper-most item and permission pairing possible.
        _item = _permission = None

        # Check study, if passed
        if study:

            # Set the item string
            item = f"{cls.ITEM}.{PPM.Study.get(study).value}"

            # Check method if necessary
            if method:
                if map.get(item) in cls.METHOD_PERMISSIONS[method]:
                    _item = item
                    _permission = map[item]
            elif map.get(item) in cls.PERMISSIONS:
                _item = item
                _permission = map[item]

        # First check permission on PPM

        # If method passed, check permission for method
        if method:
            if map.get(cls.ITEM) in cls.METHOD_PERMISSIONS[method]:
                _item = cls.ITEM
                _permission = map[cls.ITEM]

        # If no method, check for any permission
        elif map.get(cls.ITEM) in cls.PERMISSIONS:
            _item = cls.ITEM
            _permission = map[cls.ITEM]

        return _item, _permission

    @classmethod
    def has_permission(cls, permissions, method=None, study=None):
        """
        Inspects the set of permissions and returns the item
        and permission that are valid for the request. This
        returns the first found permission.

        :param permissions: A list of permissions from DBMI-AuthZ
        :type permissions: list
        :param method: The requested method to check permissions for
        :type method: str
        :param study: A specific study, defaults to None
        :type study: str, optional
        :returns: A tuple of item and permission
        :rtype: (str, str)
        """

        # Get permission
        item, permission = Auth.get_permission(permissions, method, study)

        return item is not None and permission is not None

    @classmethod
    def has_ppm_permission(cls, permissions, method=None):
        """
        Inspects the set of permissions and returns whether this
        user has absolute permission over all of PPM for the given
        operation.

        :param permissions: A list of permissions from DBMI-AuthZ
        :type permissions: list
        :param method: The requested method to check permissions for
        :type method: str
        :returns: Whether the user has the permissions or not
        :rtype: bool
        """

        # Get permission
        item, permission = Auth.get_permission(permissions, method)

        return item is not None and permission is not None

    @classmethod
    def has_study_permission(cls, permissions, study, method=None):
        """
        Inspects the set of permissions and returns whether this
        user has needed permissions on the particular PPM study
        for the given operation.

        :param permissions: A list of permissions from DBMI-AuthZ
        :type permissions: list
        :param study: A specific study
        :type study: str
        :param method: The requested method to check permissions for
        :type method: str, default None
        :returns: Whether the user has the permissions or not
        :rtype: bool
        """

        # Get permission
        item, permission = Auth.get_permission(permissions, method=method, study=study)

        return item is not None and permission is not None

    @classmethod
    def has_participant_permission(cls, permissions, participant, method=None, studies=None):
        """
        Inspects the set of permissions and returns whether this
        user has needed permissions on the particular PPM participant.
        All of the participant's studies are pulled and the requesting user is
        expected to have at least one valid permission on a study to pass.

        :param permissions: A list of permissions from DBMI-AuthZ
        :type permissions: list
        :param participant: A specific participant (email or PPM ID)
        :type participant: str
        :param method: The requested method to check permissions for
        :type method: str
        :param patient: The participant's studies, if already fetched
        :type patient: list
        :returns: Whether the user has the permissions or not
        :rtype: bool
        """
        if not studies:
            # Get participant details
            patient, studies = FHIR.query_ppm_participant_details(participant)

        # Check for study permissions on at least one
        for study in studies:

            # Run the check
            item, permission = Auth.get_permission(permissions, method, study)
            if item is not None and permission is not None:
                return True

        return False

    @classmethod
    def has_participants_permission(cls, permissions, participant_ids, method=None, participants=None):
        """
        Inspects the set of permissions and returns whether this
        user has needed permissions on the particular PPM participant.
        All of the participant's studies are pulled and the requesting user is
        expected to have at least one valid permission on a study to pass.

        :param permissions: A list of permissions from DBMI-AuthZ
        :type permissions: list
        :param participant_ids: A list of participant PPM IDs
        :type participant_ids: list
        :param method: The requested method to check permissions for
        :type method: str
        :param participants: The participant details list ([(patient, studies)]), if already fetched
        :type participants: [(dict, list)]
        :returns: Whether the user has the permissions or not
        :rtype: bool
        """
        # Get participant details
        if not participants:
            participants = FHIR.query_ppm_participants_details(participant_ids)

        # Iterate each participant
        for participant in participants:

            # Ensure we have at least a permission on a study
            studies = participant[1]
            for study in studies:

                # Run the check
                item, permission = Auth.get_permission(permissions, method, study)
                if item is not None and permission is not None:
                    break
            else:
                # Loop finished without breaking, did not have permissions on
                # at least one study
                break
        else:

            # If we are here, one permission exists on at least one study for
            # every participant in the list!
            return True

        # Participant loop was not broken early, a permission failed
        return False

    @classmethod
    def has_owner_permission(cls, email, participant, method=None, patient=None):
        """
        Inspects the requesting user's details and confirms they are
        making a request on their own data. This check if performed
        by comparing their JWT's verified email address to the stored
        email address for the Patient resource with ID matching
        the PPM/FHIR ID of the target of the request.

        :param email: The verified email address of the requesting user
        :type email: str
        :param participants: The PPM ID of the participant being requested
        :type participants: str
        :param method: The requested method to check permissions for
        :type method: str
        :param patient: The Patient object, if already fetched
        :type patient: dict
        :returns: Whether the user has the permissions or not
        :rtype: bool
        """
        # Get participant details
        if not patient:
            patient, studies = FHIR.query_ppm_participant_details(participant)

        # Compare details
        if email.lower() == patient.get("email").lower() and participant == patient.get("ppm_id"):

            # Ensure method is not DELETE
            return method is None or method != "DELETE"

        return False



import mock
import responses
import random
import unittest
import uuid
import re
import json
from datetime import datetime

from ppmutils.ppm import PPM
from ppmutils.fhir import FHIR


class TestFHIR(unittest.TestCase):
    @classmethod
    def setUpClass(cls):

        # Set the FHIR URL
        cls.fhir_url = "https://fhir.ppm.aws.dbmi-test.hms.harvard.edu"
        cls.fhir_url_pattern = r"{}/.*".format(cls.fhir_url)

        # Patch PPM FHIR URL
        cls.ppm_fhir_url_patcher = mock.patch("ppmutils.fhir.FHIR.service_url")
        cls.mock_ppm_fhir_url = cls.ppm_fhir_url_patcher.start()
        cls.mock_ppm_fhir_url.return_value = cls.fhir_url

    @classmethod
    def tearDownClass(cls):

        # Disable patcher
        cls.ppm_fhir_url_patcher.stop()

    @responses.activate
    def test_query_patient(self):

        # Set patient email
        email = "patient@email.org"

        # Build a patient with a lastname
        patient = FHIRData.patient(
            email, firstname="User", lastname="Patient", street2="Unit 500", contact_email="user@email.org",
        )
        study = FHIRData.research_study(PPM.Study.NEER)
        subject = FHIRData.research_subject("Patient/{}".format(patient["id"]), PPM.Study.NEER)
        enrollment = FHIRData.enrollment_flag("Patient/{}".format(patient["id"]))

        # Put them in a bundle
        bundle = FHIRData.create_bundle([patient, study, subject, enrollment], self.fhir_url)

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient.*"), json=bundle, status=200,
        )

        # Do the query
        queried_patient = FHIR.query_patient(email, flatten_return=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertEqual(queried_patient["email"], email)

    @responses.activate
    def test_query_patient_id(self):

        # Set patient email
        email = "patient@email.org"

        # Build a patient with a lastname
        patient = FHIRData.patient(
            email, firstname="User", lastname="Patient", street2="Unit 500", contact_email="user@email.org",
        )

        # Put them in a bundle
        bundle = FHIRData.create_bundle([patient], self.fhir_url)

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient.*"), json=bundle, status=200,
        )

        # Do the query
        ppm_id = FHIR.query_patient_id(email)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertEqual(ppm_id, patient["id"])

    @responses.activate
    def test_query_participant(self):

        # Set patient email
        email = "patient@email.org"

        # Build a patient with a lastname
        patient = FHIRData.patient(
            email, firstname="User", lastname="Patient", street2="Unit 500", contact_email="user@email.org",
        )
        study = FHIRData.research_study(PPM.Study.NEER)
        subject = FHIRData.research_subject("Patient/{}".format(patient["id"]), PPM.Study.NEER)
        enrollment = FHIRData.enrollment_flag("Patient/{}".format(patient["id"]))

        # Put them in a bundle
        bundle = FHIRData.create_bundle([patient, study, subject, enrollment], self.fhir_url)

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient.*"), json=bundle, status=200,
        )

        # Do the query
        queried_patient = FHIR.query_participant(email, flatten_return=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertEqual(queried_patient["email"], email)
        self.assertEqual(queried_patient["study"], PPM.Study.NEER.value)
        self.assertEqual(queried_patient["enrollment"], PPM.Enrollment.Registered.value)

    @responses.activate
    def test_query_enrollment_status(self):

        # Set patient email
        email = "patient@email.org"

        # Build a patient with a lastname
        patient = FHIRData.patient(
            email, firstname="User", lastname="Patient", street2="Unit 500", contact_email="user@email.org",
        )
        study = FHIRData.research_study(PPM.Study.NEER)
        subject = FHIRData.research_subject("Patient/{}".format(patient["id"]), PPM.Study.NEER)
        enrollment = FHIRData.enrollment_flag("Patient/{}".format(patient["id"]), PPM.Enrollment.Proposed.value)

        # Put them in a bundle
        bundle = FHIRData.create_bundle([patient, study, subject, enrollment], self.fhir_url)

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Flag.*"), json=bundle, status=200,
        )

        # Do the query
        queried_enrollment = FHIR.query_enrollment_status(email)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(queried_enrollment == PPM.Enrollment.Proposed.value)

    @responses.activate
    def test_query_participants(self):

        # Start a data set
        data = FHIRData(self.fhir_url, participants=3)

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient.*"), json=data.bundle, status=200,
        )

        # Do the query
        participants = FHIR.query_participants(testing=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertEqual(len(participants), 3)

        # Check some properties
        self.assertTrue(PPM.Study.enum(participants[2]["study"]) in PPM.Study)
        self.assertTrue(participants[1]["email"] is not None)
        self.assertTrue(PPM.Enrollment.enum(participants[2]["enrollment"]) in PPM.Enrollment)

    @responses.activate
    def test_patient_update_lastname(self):

        # Build a patient with a lastname
        patient = FHIRData.patient("patient@email.org", firstname="User", lastname="Patient")

        # Set an example form
        form = {"firstname": "Newer", "lastname": None}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertTrue(payload["name"][0].get("family", False))

    @responses.activate
    def test_patient_update_add_address2(self):

        # Build a patient with a lastname
        patient = FHIRData.patient("patient@email.org", firstname="User", lastname="Patient")

        # Set an example form
        form = {"street_address1": "3100 Some Address", "street_address2": "Unit 401"}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertTrue(len(payload["address"][0]["line"]) > 1)

    @responses.activate
    def test_patient_update_remove_address2(self):

        # Build a patient with a lastname
        patient = FHIRData.patient("patient@email.org", firstname="User", lastname="Patient", street2="Unit 500",)

        # Set an example form
        form = {"street_address1": "3100 Some Address", "street_address2": None}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertTrue(len(payload["address"][0]["line"]) == 1)

    @responses.activate
    def test_patient_update_add_contact_email(self):

        # Build a patient with a lastname
        patient = FHIRData.patient("patient@email.org", firstname="User", lastname="Patient", street2="Unit 500",)

        # Set an example form
        form = {"contact_email": "user@email.org"}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertTrue(next((telecom for telecom in payload["telecom"] if telecom["system"] == "email"), False,))

    @responses.activate
    def test_patient_update_contact_email(self):

        # Set the new email
        contact_email = "somenewemail@email.com"

        # Build a patient with a lastname
        patient = FHIRData.patient("patient@email.org", firstname="User", lastname="Patient", street2="Unit 500",)

        # Set an example form
        form = {"contact_email": contact_email}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertEqual(
            next(telecom for telecom in payload["telecom"] if telecom["system"] == "email")["value"], contact_email,
        )

    @responses.activate
    def test_patient_update_remove_contact_email(self):

        # Build a patient with a lastname
        patient = FHIRData.patient(
            "patient@email.org",
            firstname="User",
            lastname="Patient",
            street2="Unit 500",
            contact_email="user@email.org",
        )

        # Set an example form
        form = {"contact_email": None}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertFalse(next((telecom for telecom in payload["telecom"] if telecom["system"] == "email"), False,))

    @responses.activate
    def test_patient_update_add_referral(self):

        # Build a patient with a lastname
        patient = FHIRData.patient("patient@email.org", firstname="User", lastname="Patient", street2="Unit 500",)

        # Set an example form
        form = {"how_did_you_hear_about_us": "I was referred by John Smith"}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertTrue(next((e for e in payload["extension"] if e["url"] == FHIR.referral_extension_url), False,))

    @responses.activate
    def test_patient_update_referral(self):

        # Set new value
        referral = "new_referral"

        # Build a patient with a lastname
        patient = FHIRData.patient("patient@email.org", firstname="User", lastname="Patient", street2="Unit 500",)
        patient.setdefault("extension", []).append({"url": FHIR.referral_extension_url, "valueString": "old_referral"})

        # Set an example form
        form = {"how_did_you_hear_about_us": referral}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertEqual(
            next((e for e in payload["extension"] if e["url"] == FHIR.referral_extension_url), False,)["valueString"],
            referral,
        )

    @responses.activate
    def test_patient_update_remove_referral(self):

        # Build a patient with a lastname
        patient = FHIRData.patient(
            "patient@email.org",
            firstname="User",
            lastname="Patient",
            street2="Unit 500",
            contact_email="user@email.org",
        )

        # Set an example form
        form = {"how_did_you_hear_about_us": None}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure last name was removed
        self.assertFalse(next((e for e in payload["extension"] if e["url"] == FHIR.referral_extension_url), False,))

    @responses.activate
    def test_patient_update_requirements(self):

        # Build a patient with a lastname
        patient = FHIRData.patient(
            "patient@email.org",
            firstname="User",
            lastname="Patient",
            street2="Unit 500",
            contact_email="user@email.org",
        )

        # Set an example form
        form = {"firstname": None, "email": None, "city": None, "phone": None}

        # Build the response handler
        responses.add(
            responses.GET, re.compile(self.fhir_url + r"/Patient/.*"), json=patient, status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient/.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient(patient["id"], form=form)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure properties still exist
        self.assertTrue(len(payload["name"][0]["given"]) > 0)
        self.assertTrue(
            next((id["value"] for id in payload["identifier"] if id["system"] == "http://schema.org/email"), False,)
        )
        self.assertTrue(payload["address"][0].get("city", False))
        self.assertTrue(
            next((telecom["value"] for telecom in payload["telecom"] if telecom["system"] == "phone"), False,)
        )

    @responses.activate
    def test_update_patient_deceased_1(self):

        # Set the deceased date
        deceased = datetime.now()

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PATCH,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_deceased(patient["id"], deceased)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[0].request.body)

        # Ensure properties still exist
        self.assertEqual(payload[0]["op"], "replace")
        self.assertEqual(payload[0]["value"], deceased.isoformat())

    @responses.activate
    def test_update_patient_deceased_2(self):

        # Set the deceased date
        deceased = None

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PATCH,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_deceased(patient["id"], deceased)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[0].request.body)

        # Ensure properties still exist
        self.assertEqual(payload[0]["op"], "remove")

    @responses.activate
    def test_update_patient_deceased_3(self):

        # Set the deceased date
        deceased = datetime.now()

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PATCH,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_deceased(patient["id"], deceased, active=False)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[0].request.body)

        # Ensure properties still exist
        self.assertEqual(payload[0]["op"], "replace")
        self.assertEqual(payload[0]["path"], "/deceasedDateTime")
        self.assertEqual(payload[0]["value"], deceased.isoformat())
        self.assertEqual(payload[1]["op"], "replace")
        self.assertEqual(payload[1]["path"], "/active")
        self.assertEqual(payload[1]["value"], False)

    @responses.activate
    def test_update_patient_deceased_4(self):

        # Set the deceased date
        deceased = None

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PATCH,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_deceased(patient["id"], deceased, active=True)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[0].request.body)

        # Ensure properties still exist
        self.assertEqual(payload[0]["op"], "remove")
        self.assertEqual(payload[0]["path"], "/deceasedDateTime")
        self.assertEqual(payload[1]["op"], "replace")
        self.assertEqual(payload[1]["path"], "/active")
        self.assertEqual(payload[1]["value"], True)

    @responses.activate
    def test_update_patient_twitter_1(self):

        # Set the the handle
        handle = "sometwitterhandle"
        uses_twitter = None

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_twitter(patient["id"], handle, uses_twitter)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Get telecom
        telecom = next(t for t in payload["telecom"] if t["system"] == FHIR.patient_twitter_telecom_system)
        self.assertEqual(telecom["value"], "https://twitter.com/" + handle)

        # Get extension
        extension = next(e for e in payload["extension"] if e["url"] == FHIR.twitter_extension_url)
        self.assertEqual(extension["valueBoolean"], True)

    @responses.activate
    def test_update_patient_twitter_2(self):

        # Set the the handle
        handle = None
        uses_twitter = None

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_twitter(patient["id"], handle, uses_twitter)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Get telecom
        telecom = next((t for t in payload["telecom"] if t["system"] == FHIR.patient_twitter_telecom_system), None,)
        self.assertIsNone(telecom)

        # Get extension
        extension = next((e for e in payload["extension"] if e["url"] == FHIR.twitter_extension_url), None,)
        self.assertIsNone(extension)

    @responses.activate
    def test_update_patient_twitter_3(self):

        # Set the the handle
        handle = None
        uses_twitter = False

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_twitter(patient["id"], handle, uses_twitter)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Get telecom
        telecom = next((t for t in payload["telecom"] if t["system"] == FHIR.patient_twitter_telecom_system), None,)
        self.assertIsNone(telecom)

        # Get extension
        extension = next((e for e in payload["extension"] if e["url"] == FHIR.twitter_extension_url), None,)
        self.assertEqual(extension["valueBoolean"], uses_twitter)

    @responses.activate
    def test_update_patient_twitter_4(self):

        # Set the the handle
        handle = None
        uses_twitter = False

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Set an initial handle and extension
        patient["telecom"].append(
            {"system": FHIR.patient_twitter_telecom_system, "value": "https://twitter.com/somehandle",}
        )
        patient["extension"].append({"url": FHIR.twitter_extension_url, "valueBoolean": True})

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        responses.add(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            json={"operation": "Patient resource updated!"},
            status=200,
        )

        # Do the update
        updated = FHIR.update_twitter(patient["id"], handle, uses_twitter)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        # Check request
        payload = json.loads(responses.calls[1].request.body)

        # Get telecom
        telecom = next((t for t in payload["telecom"] if t["system"] == FHIR.patient_twitter_telecom_system), None,)
        self.assertIsNone(telecom)

        # Get extension
        extension = next((e for e in payload["extension"] if e["url"] == FHIR.twitter_extension_url), None,)
        self.assertEqual(extension["valueBoolean"], uses_twitter)

    @responses.activate
    def test_update_patient_extension_1(self):

        # Set the the extension details
        url = FHIR.facebook_extension_url
        value = False

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_extension(patient["id"], url, value)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Get extension
        extension = next((e for e in payload["extension"] if e["url"] == url), None)
        self.assertEqual(extension["valueBoolean"], value)

    @responses.activate
    def test_update_patient_extension_2(self):

        # Set the the extension details
        url = FHIR.facebook_extension_url
        value = False

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Set an initial handle and extension
        patient["extension"].append({"url": url, "valueBoolean": not value})

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_extension(patient["id"], url, value)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Get extension
        extension = next((e for e in payload["extension"] if e["url"] == url), None)
        self.assertEqual(extension["valueBoolean"], value)

    @responses.activate
    def test_update_patient_extension_3(self):

        # Set the the extension details
        url = FHIR.facebook_extension_url
        value = "somestring"

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Set an initial handle and extension
        patient["extension"].append({"url": url, "valueString": "someotherstring"})

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_extension(patient["id"], url, value)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Get extension
        extension = next((e for e in payload["extension"] if e["url"] == url), None)
        self.assertEqual(extension["valueString"], value)

    @responses.activate
    def test_update_patient_picnichealth_1(self):

        # Set the the extension details
        url = FHIR.picnichealth_extension_url
        value = False

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_picnichealth(patient["id"], value)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Get extension
        extension = next((e for e in payload["extension"] if e["url"] == url), None)
        self.assertEqual(extension["valueBoolean"], value)

    @responses.activate
    def test_update_patient_picnichealth_2(self):

        # Set the the extension details
        url = FHIR.picnichealth_extension_url
        value = True

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Set an initial handle and extension
        patient["extension"].append({"url": url, "valueBoolean": not value})

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Patient resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Patient.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_picnichealth(patient["id"], value)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Get extension
        extension = next((e for e in payload["extension"] if e["url"] == url), None)
        self.assertEqual(extension["valueBoolean"], value)

    @responses.activate
    def test_update_enrollment_1(self):

        # Set the enrollment value we are setting
        enrollment = PPM.Enrollment.Consented.value

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Flag.*"),
            json=FHIRData.create_bundle([flag], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Flag.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_enrollment(patient["id"], status=enrollment)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure properties still exist
        self.assertEqual(payload["subject"]["reference"], "Patient/{}".format(patient["id"]))
        self.assertEqual(payload["code"]["coding"][0]["code"], enrollment)
        self.assertEqual(payload["code"]["text"], PPM.Enrollment.title(enrollment))
        self.assertEqual(payload["status"], "inactive")
        self.assertIsNone(payload["period"].get("end"))

    @responses.activate
    def test_update_enrollment_2(self):

        # Set the enrollment value we are setting
        enrollment = PPM.Enrollment.Proposed.value

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Flag.*"),
            json=FHIRData.create_bundle([flag], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Flag.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_enrollment(patient["id"], status=enrollment)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure properties still exist
        self.assertEqual(payload["subject"]["reference"], "Patient/{}".format(patient["id"]))
        self.assertEqual(payload["code"]["coding"][0]["code"], enrollment)
        self.assertEqual(payload["code"]["text"], PPM.Enrollment.title(enrollment))
        self.assertNotEqual(payload["status"], "active")
        self.assertIsNone(payload["period"].get("end"))

    @responses.activate
    def test_update_enrollment_3(self):

        # Set the enrollment value we are setting
        enrollment = PPM.Enrollment.Accepted.value

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Flag.*"),
            json=FHIRData.create_bundle([flag], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Flag.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_enrollment(patient["id"], status=enrollment)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure properties still exist
        self.assertEqual(payload["subject"]["reference"], "Patient/{}".format(patient["id"]))
        self.assertEqual(payload["code"]["coding"][0]["code"], enrollment)
        self.assertEqual(payload["code"]["text"], PPM.Enrollment.title(enrollment))
        self.assertEqual(payload["status"], "active")
        self.assertIsNotNone(payload["period"].get("start"))
        self.assertIsNone(payload["period"].get("end"))

    @responses.activate
    def test_update_enrollment_4(self):

        # Set the enrollment value we are setting
        enrollment = PPM.Enrollment.Terminated.value

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Flag.*"),
            json=FHIRData.create_bundle([flag], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Flag.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_enrollment(patient["id"], status=enrollment)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure properties still exist
        self.assertEqual(payload["subject"]["reference"], "Patient/{}".format(patient["id"]))
        self.assertEqual(payload["code"]["coding"][0]["code"], enrollment)
        self.assertEqual(payload["code"]["text"], PPM.Enrollment.title(enrollment))
        self.assertEqual(payload["status"], "inactive")
        self.assertIsNotNone(payload["period"].get("start"))
        self.assertIsNotNone(payload["period"].get("end"))

    @responses.activate
    def test_update_enrollment_5(self):

        # Set the enrollment value we are setting
        enrollment = PPM.Enrollment.Completed.value

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Flag.*"),
            json=FHIRData.create_bundle([flag], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PUT,
            re.compile(self.fhir_url + r"/Flag.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_patient_enrollment(patient["id"], status=enrollment)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure properties still exist
        self.assertEqual(payload["subject"]["reference"], "Patient/{}".format(patient["id"]))
        self.assertEqual(payload["code"]["coding"][0]["code"], enrollment)
        self.assertEqual(payload["code"]["text"], PPM.Enrollment.title(enrollment))
        self.assertEqual(payload["status"], "inactive")
        self.assertIsNotNone(payload["period"].get("start"))
        self.assertIsNotNone(payload["period"].get("end"))

    @responses.activate
    def test_update_research_subject_1(self):

        # Set the dates
        end = datetime.now()

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PATCH,
            re.compile(self.fhir_url + r"/ResearchSubject.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_research_subject(patient["id"], research_subject["id"], end=end)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[0].request.body)

        # Ensure properties still exist
        self.assertEqual(payload[0]["op"], "add")
        self.assertEqual(payload[0]["value"], end.isoformat())

    @responses.activate
    def test_update_research_subject_2(self):

        # Set the dates
        end = None

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PATCH,
            re.compile(self.fhir_url + r"/ResearchSubject.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_research_subject(patient["id"], research_subject["id"], end=end)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[0].request.body)

        # Ensure properties still exist
        self.assertEqual(payload[0]["op"], "remove")

    @responses.activate
    def test_update_ppm_research_subject_1(self):

        # Set the dates
        end = datetime.now()

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/ResearchSubject.*"),
            json=FHIRData.create_bundle([flag], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PATCH,
            re.compile(self.fhir_url + r"/ResearchSubject.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_ppm_research_subject(patient["id"], PPM.Study.NEER.value, end=end)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure properties still exist
        self.assertEqual(payload[0]["op"], "add")
        self.assertEqual(payload[0]["value"], end.isoformat())

    @responses.activate
    def test_update_ppm_research_subject_2(self):

        # Set the dates
        end = None

        # Start a data set
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/ResearchSubject.*"),
            json=FHIRData.create_bundle([flag], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Flag resource updated!"})

        responses.add_callback(
            responses.PATCH,
            re.compile(self.fhir_url + r"/ResearchSubject.*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        updated = FHIR.update_ppm_research_subject(patient["id"], PPM.Study.NEER.value, end=end)

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)
        self.assertTrue(updated)

        payload = json.loads(responses.calls[1].request.body)

        # Ensure properties still exist
        self.assertEqual(payload[0]["op"], "remove")

    @responses.activate
    def test_delete_participant(self):

        # Create our participant to be deleted
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        # Collect IDs
        resource_ids = [
            "Patient/" + patient["id"],
            "Flag/" + flag["id"],
            "ResearchSubject/" + research_subject["id"],
        ]

        # Build the response handler
        responses.add(
            responses.GET,
            re.compile(self.fhir_url + r"/Patient.*"),
            json=FHIRData.create_bundle([patient, flag, research_subject, research_study], self.fhir_url),
            status=200,
        )

        def update_callback(request):

            return 200, {}, json.dumps({"operation": "Resources were deleted!"})

        responses.add_callback(
            responses.POST,
            re.compile(self.fhir_url + r".*"),
            callback=update_callback,
            content_type="application/json",
        )

        # Do the update
        FHIR.delete_participant(patient["id"])

        # Check it
        self.assertGreaterEqual(len(responses.calls), 2)

        # Check each transaction request
        payload = json.loads(responses.calls[1].request.body)
        for request in [e["request"] for e in payload["entry"]]:

            # Ensure it only relates to our single participant
            self.assertIn(request["url"], resource_ids)
            self.assertEqual(request["method"], "DELETE")

            # Ensure the PPM ResearchStudy was unharmed
            self.assertNotEqual(request["url"], "ResearchStudy/" + research_study["id"])

    @responses.activate
    def test_delete_patient(self):

        # Create our participant to be deleted
        research_study, patient, flag, research_subject = FHIRData.participant(PPM.Study.NEER)

        def delete_callback(request):

            return 200, {}, json.dumps({"operation": "Resource was deleted!"})

        responses.add_callback(
            responses.DELETE,
            re.compile(self.fhir_url + r".*"),
            callback=delete_callback,
            content_type="application/json",
        )

        # Do the update
        FHIR._delete_resource("Patient", patient["id"])

        # Check it
        self.assertGreaterEqual(len(responses.calls), 1)

        # Check each transaction request
        self.assertIn("Patient/" + patient["id"], responses.calls[0].request.url)


class FHIRData(object):
    """This class is used to manage the emulated data set from which to test
    PPM FHIR methods"""

    bundle = None
    fhir_url = None

    def __init__(self, fhir_url, participants=1, study=None):

        # Retain
        self.fhir_url = fhir_url

        # Create participants and create our bundle
        for _ in range(0, participants):

            # If not study, randomize it
            if not study:
                study = PPM.Study(
                    list(dict(PPM.Study.choices()).keys())[random.randint(1, len(PPM.Study.choices())) - 1]
                )

            # Create the resources
            (research_study, patient, enrollment, research_subject,) = FHIRData.participant(study)

            # Add the study if not already there
            if not self.bundle:

                # Start it
                self.bundle = FHIRData.create_bundle([research_study], fhir_url)

            elif not next(r for r in self.bundle["entry"] if r["resource"]["id"] == research_study["id"]):

                # Add it
                self.bundle["entry"].extend(FHIRData.create_bundle([research_study], fhir_url)["entry"])

            # Add remaining resources
            self.bundle["entry"].extend(
                FHIRData.create_bundle([patient, enrollment, research_subject], fhir_url)["entry"]
            )

    def add_participant(self, patient, enrollment, research_subject, research_study):
        """
        Adds the participant comprised of the needed resources to the Bundle resource.
        :param patient: The Patient
        :param enrollment: The enrollment Flag
        :param research_subject: The ResearchSubject
        :param research_study: The ResearchStudy
        :return: None
        """
        # Add the study if not already there
        if not self.bundle:

            # Start it
            self.bundle = FHIRData.create_bundle([research_study], self.fhir_url)

        elif not next((r for r in self.bundle["entry"] if r["resource"]["id"] == research_study["id"]), None,):

            # Add it
            self.bundle["entry"].extend(FHIRData.create_bundle([research_study], self.fhir_url)["entry"])

        # Add remaining resources
        self.bundle["entry"].extend(
            FHIRData.create_bundle([patient, enrollment, research_subject], self.fhir_url)["entry"]
        )

    @staticmethod
    def participant(study, enrollment=PPM.Enrollment.Registered):
        """
        Builds and returns a bundle comprised of n participant resource sets.
        This includes Patient, Flag, ResearchSubject and ResearchStudy and others,
        if requested.
        :param study: The study the participant should belong to
        :param enrollment: The enrollment status they should be set to
        :return: A tuple of FHIR resource dicts
        """
        # Make a random ID
        ppm_id = f"{random.randint(1, 99999)}"

        # Set patient email
        email = f"patient-{ppm_id}-{PPM.Study.enum(study).value}@email.org"

        # Build a patient with a lastname
        patient = FHIRData.patient(
            email,
            firstname=f"User {ppm_id}",
            lastname="Patient",
            street1=f"{random.randint(1000, 9999)} Main St.",
            street2="Unit 500",
            city="Boston",
            state="Massachusetts",
            zip=f"{random.randint(2000, 20000)}",
            contact_email=f"patient-{ppm_id}-{PPM.Study.enum(study).value}@email.org",
        )
        research_study = FHIRData.research_study(study)
        research_subject = FHIRData.research_subject("Patient/{}".format(patient["id"]), study)
        enrollment_flag = FHIRData.enrollment_flag("Patient/{}".format(patient["id"]), enrollment.value)

        return research_study, patient, enrollment_flag, research_subject

    @staticmethod
    def create_bundle(resources, fhir_url):
        """
        Builds and returns a JSON Bundle resource containing the passed JSON
        FHIR resources
        :param resources: A list of FHIR resource dicts
        :return: dict
        """
        return {
            "resourceType": "Bundle",
            "id": f"{uuid.uuid4()}",
            "meta": {"lastUpdated": datetime.now().isoformat()},
            "type": "searchset",
            "total": len(resources),
            "link": [{"relation": "self", "url": fhir_url}],
            "entry": [
                {"resource": resource, "fullUrl": f'{fhir_url}/{resource["resourceType"]}/' f'{resource["id"]}',}
                for resource in resources
            ],
        }

    @staticmethod
    def patient(
        email,
        firstname,
        lastname=None,
        street1="49 West New Drive",
        street2=None,
        city="Boston",
        state="MA",
        zip="02445",
        phone="+1 (353) 233-2708",
        contact_email="user@email.org",
        twitter=None,
        identifier=str(uuid.uuid4()),
    ):
        """
        Initializes and returns a test Patient resource using the passed parameters.
        Required items not passed are randomly generated, non-required items are
        left blank.
        :return: The Patient FHIR resource as a dict
        :rtype: dict
        """
        # Convert to form
        form = {
            "firstname": firstname,
            "lastname": lastname,
            "email": email,
            "street_address1": street1,
            "street_address2": street2,
            "city": city,
            "state": state,
            "zip": zip,
            "phone": phone,
            "contact_email": contact_email,
            "how_did_you_hear_about_us": "Example of how they heard about PPM",
            "twitter_handle": twitter,
        }

        # Get from FHIR.Resource
        data = FHIR.Resources.patient(form)

        # Replace/set some additional data
        data["id"] = f"{random.randint(1, 9999)}"
        data["identifier"].append(
            {"system": "https://peoplepoweredmedicine.org/fhir/patient", "value": identifier,}
        )

        # Toggle Twitter usage
        data["extension"].append(
            {
                "url": "https://p2m2.dbmi.hms.harvard.edu/fhir/" "StructureDefinition/uses-twitter",
                "valueBoolean": twitter is not None,
            }
        )

        return data

    @staticmethod
    def research_study(study):
        """
        Initializes and returns a test ResearchSubject resource using the passed
        parameters. Required items not passed are randomly generated, non-required
        items are left blank.
        :return: The ResearchSubject FHIR resource as a dict
        :rtype: dict
        """
        # Ensure we pass a study value
        study = PPM.Study.enum(study).value

        # Build initial object
        data = FHIR.Resources.ppm_research_study(study, PPM.Study.title(study))

        return data

    @staticmethod
    def research_subject(patient, study, start=None, end=None):
        """
        Initializes and returns a test ResearchSubject resource using the passed
        parameters. Required items not passed are randomly generated,
        non-required items are left blank.
        :return: The ResearchSubject FHIR resource as a dict
        :rtype: dict
        """
        # Ensure we pass a study value
        study = PPM.Study.enum(study).value

        # Build initial object
        data = FHIR.Resources.ppm_research_subject(study, patient)

        # Set random id
        data["id"] = f"{random.randint(1, 9999)}"

        # Check dates
        if start:
            data["period"]["start"] = start

        if end:
            data["period"]["end"] = end

        return data

    @staticmethod
    def enrollment_flag(
        patient, enrollment=PPM.Enrollment.Registered.value, start=datetime.now(), end=None,
    ):
        """
        Initializes and returns a test ResearchSubject resource using the
        passed parameters. Required items not passed are randomly generated,
        non-required items are left blank.
        :return: The ResearchSubject FHIR resource as a dict
        :rtype: dict
        """
        # Build initial object
        data = FHIR.Resources.enrollment_flag(patient, status=enrollment, start=start, end=end)

        # Set random id
        data["id"] = f"{random.randint(1, 9999)}"

        return data



from ppmutils.ppm import PPM
from ppmutils.settings import ppm_settings

# Get the app logger
import logging

logger = logging.getLogger(ppm_settings.LOGGER_NAME)


class Participants(PPM.Service):

    service = "participants"
    proxied = True

    @classmethod
    def get_participants(cls, request, study=None, enrollments=None, fhir=None):
        """
        Returns a list of all participants
        """
        # Build query
        query = {}
        if study:
            query["studies"] = study

        if enrollments:
            query["enrollments"] = ",".join(enrollments)

        if fhir is not None:
            query["fhir"] = True

        # Make the request
        return cls.get(request=request, path=f"/", data=query)

    @classmethod
    def get_participant(cls, request, patient_ref, fhir=None):
        """
        Gets a single participant for the passed ID
        """
        query = {}
        if fhir is not None:
            query["fhir"] = True

        # Make the request
        return cls.get(request=request, path=f"/{patient_ref}/", data=query)

    @classmethod
    def create_participant(cls, request, form):
        """
        Deletes the participant and their record
        """
        # Make the request
        return cls.post(request=request, path=f"/", data=form)

    @classmethod
    def update_participant(cls, request, patient_ref, form):
        """
        Performs an update operation on the participant
        """
        # Make the request
        return cls.post(request=request, path=f"/{patient_ref}", data=form)



import mock
import unittest
from furl import furl
from ppmutils.ppm import PPM

# Set test data
SERVICE_URL = "https://some.test.api.address/"


class TestPPMService(unittest.TestCase):
    @mock.patch("ppmutils.ppm.PPM.Service.service_url")
    def test_url_build_1(self, mock_service_url):

        # Set the service URL to use
        mock_service_url.return_value = SERVICE_URL

        # Set an example path
        path = "/some/example/path"

        # Build a URL
        url = furl(PPM.Service._build_url(path=path))

        # Check it
        self.assertTrue("//" not in str(url.path))
        self.assertFalse(str(url.query))
        self.assertEqual("some.test.api.address", url.netloc)

    @mock.patch("ppmutils.ppm.PPM.Service.service_url")
    def test_url_build_2(self, mock_service_url):

        # Set the service URL to use
        mock_service_url.return_value = SERVICE_URL

        # Set an example path
        path = "/some/example/path?with=query&string=included"

        # Build a URL
        url = furl(PPM.Service._build_url(path=path))

        # Check it
        self.assertTrue("//" not in str(url.path))
        self.assertEqual(len(url.query.params.items()), 2)
        self.assertEqual("some.test.api.address", url.netloc)

    @mock.patch("ppmutils.ppm.PPM.Service.service_url")
    def test_url_build_3(self, mock_service_url):

        # Set the service URL to use
        mock_service_url.return_value = SERVICE_URL

        # Set an example path
        path = "/some//example//path"

        # Build a URL
        url = furl(PPM.Service._build_url(path=path))

        # Check it
        self.assertTrue("//" not in str(url.path))
        self.assertFalse(str(url.query))
        self.assertEqual("some.test.api.address", url.netloc)

    @mock.patch("ppmutils.ppm.PPM.Service.service_url")
    def test_url_build_4(self, mock_service_url):

        # Set the service URL to use
        mock_service_url.return_value = SERVICE_URL

        # Set an example path
        path = "/some//example//path?and=query&params=included+12"

        # Build a URL
        url = furl(PPM.Service._build_url(path=path))

        # Check it
        self.assertTrue("//" not in str(url.path))
        self.assertEqual(len(url.query.params.items()), 2)
        self.assertEqual("some.test.api.address", url.netloc)

    def test_ppm_study_meta_1(self):

        # Compare methods for determining study meta
        self.assertEqual(PPM.Study.title(PPM.Study.NEER), PPM.Study.title(PPM.Study.NEER.value))
        self.assertEqual(PPM.Study.title(PPM.Study.ASD), PPM.Study.title(PPM.Study.ASD.value))

        # Use FHIR codes
        self.assertEqual(
            PPM.Study.title(PPM.Study.ASD), PPM.Study.title(PPM.Study.fhir_id(PPM.Study.ASD)),
        )
        self.assertEqual(
            PPM.Study.title(PPM.Study.ASD), PPM.Study.title(PPM.Study.fhir_id(PPM.Study.ASD.value)),
        )

    def test_ppm_study_meta_2(self):

        # Compare methods for determining study meta
        self.assertTrue(PPM.Study.get(PPM.Study.NEER) is PPM.Study.NEER)
        self.assertTrue(PPM.Study.get(PPM.Study.NEER.value) is PPM.Study.NEER)
        self.assertTrue(PPM.Study.get(PPM.Study.ASD) is PPM.Study.ASD)
        self.assertTrue(PPM.Study.get(PPM.Study.ASD.value) is PPM.Study.ASD)

    def test_ppm_study_meta_3(self):

        # Compare methods for determining study meta
        self.assertTrue(PPM.Study.is_ppm(PPM.Study.fhir_id(PPM.Study.NEER)))
        self.assertTrue(PPM.Study.is_ppm(PPM.Study.fhir_id(PPM.Study.ASD)))

    def test_ppm_study_meta_4(self):

        # Compare methods for determining study meta
        ppm_enum = PPM.Study
        for enum in ppm_enum:
            # Compare getters on name
            self.assertEqual(ppm_enum.enum(enum.name), enum)
            # Compare getters on value
            self.assertEqual(ppm_enum.enum(enum.value), enum)
            # Compare getters on title
            self.assertEqual(ppm_enum.enum(ppm_enum.title(enum)), enum)
            # Compare getters on identifier
            self.assertEqual(ppm_enum.enum(ppm_enum.fhir_id(enum)), enum)

        # Check edge cases
        self.assertEqual(ppm_enum.enum("ppm-asd"), PPM.Study.ASD)
        self.assertEqual(ppm_enum.enum("asd"), PPM.Study.ASD)

    def test_ppm_enrollment_meta_1(self):

        # Compare methods for determining study meta
        ppm_enum = PPM.Enrollment
        for enum in ppm_enum:
            # Compare getters on name
            self.assertEqual(ppm_enum.enum(enum.name), enum)
            # Compare getters on value
            self.assertEqual(ppm_enum.enum(enum.value), enum)
            # Compare getters on title
            self.assertEqual(ppm_enum.enum(ppm_enum.title(enum)), enum)

    def test_ppm_communication_meta_1(self):

        # Compare methods for determining study meta
        ppm_enum = PPM.Communication
        for enum in ppm_enum:
            # Compare getters on name
            self.assertEqual(ppm_enum.enum(enum.name), enum)
            # Compare getters on value
            self.assertEqual(ppm_enum.enum(enum.value), enum)
            # Compare getters on title
            self.assertEqual(ppm_enum.enum(ppm_enum.title(enum)), enum)

    def test_ppm_provider_meta_1(self):

        # Compare methods for determining study meta
        ppm_enum = PPM.Provider
        for enum in ppm_enum:
            # Compare getters on name
            self.assertEqual(ppm_enum.enum(enum.name), enum)
            # Compare getters on value
            self.assertEqual(ppm_enum.enum(enum.value), enum)
            # Compare getters on title
            self.assertEqual(ppm_enum.enum(ppm_enum.title(enum)), enum)

    def test_ppm_tracked_item_meta_1(self):

        # Compare methods for determining study meta
        ppm_enum = PPM.TrackedItem
        for enum in ppm_enum:
            # Compare getters on name
            self.assertEqual(ppm_enum.enum(enum.name), enum)
            # Compare getters on value
            self.assertEqual(ppm_enum.enum(enum.value), enum)
            # Compare getters on title
            self.assertEqual(ppm_enum.enum(ppm_enum.title(enum)), enum)



from enum import Enum
import requests
from furl import furl
import json
import re
import os
from functools import total_ordering

from django.template.loader import render_to_string
from django.core.mail import EmailMultiAlternatives
from django.conf import settings
from ppmutils.settings import ppm_settings

# Get the app logger
import logging

logger = logging.getLogger(ppm_settings.LOGGER_NAME)


# Adding this decorator to an Enum allows it to
# be passed into a Django context and its members
# accessed via name and value as well as being iterated (untested)
def django_enum(cls):
    cls.do_not_call_in_templates = True
    return cls


class PPMEnum(Enum):
    """
    An extended Enum class with some convenience methods for working with
    enum values/keys/etc
    """

    @classmethod
    def enum(cls, enum):
        """Accepts any form of an enum and returns the enum"""
        for item in cls:
            if enum is item or enum == item.name or enum == item.value:
                return item

            # Compare titles
            if item.value in dict(cls.choices()) and dict(cls.choices())[item.value] == enum:
                return item

        raise ValueError('Value "{}" is not a valid {}'.format(enum, cls.__name__))

    @classmethod
    def get(cls, enum):
        """
        Returns an instance of the Study enum for the given study value, enum, whatever
        :param enum: The study value string, name or enum
        :type enum: Object
        :return: The instance of PPM Study enum
        :rtype: PPM.Study
        """
        return cls.enum(enum)

    @classmethod
    def from_value(cls, value):
        """
        Returns an instance of the Study enum for the given study value
        :param study: The study value string
        :type study: str
        :return: The instance of PPM Study enum
        :rtype: PPM.Study
        """
        return cls.enum(value)

    @classmethod
    def title(cls, enum):
        """
        Returns the title to be used for the given enum.
        :param enum: The enum identifier/name/value
        :type enum: object
        :return: The title for the enum
        :rtype: str
        """
        # Get the value
        value = cls.get(enum).value

        # Try choices
        return dict(cls.choices())[value] if value in dict(cls.choices()) else cls.get(enum).name

    @classmethod
    def choices(cls):
        """
        Returns a choices tuple of tuples. Define enum titles if different
        from names/values here as this is the source for pulling enum titles.
        :return: ((str, str), )
        """
        return tuple((e.name, e.value) for e in cls)


class PPM:
    """
    This class serves mostly to track the PPM project properties across
    studies and consolidate functionality common amongst all services.
    """

    # Set values for determining environments
    class Environment(PPMEnum):
        Local = "local"
        Dev = "dev"
        Staging = "staging"
        Prod = "prod"

    @staticmethod
    def is_tester(email):
        """
        Checks test user email patterns and returns True if a user's email
        matches.
        :param email: The user's email address
        :return: bool
        """
        if hasattr(settings, "TEST_EMAIL_PATTERNS") and type(getattr(settings, "TEST_EMAIL_PATTERNS")) is str:
            testers = settings.TEST_EMAIL_PATTERNS.split(",")
        elif hasattr(settings, "TEST_EMAIL_PATTERNS") and type(getattr(settings, "TEST_EMAIL_PATTERNS")) is list:
            testers = settings.TEST_EMAIL_PATTERNS
        else:
            return False

        # Iterate through all patterns
        for pattern in testers:
            if re.match(pattern, email):
                return True

        return False

    @django_enum
    class Study(PPMEnum):
        NEER = "neer"
        ASD = "asd"
        EXAMPLE = "example"
        RANT = "rant"

        @staticmethod
        def equals(this, that):
            """
            Compares a reference to a study and returns whether it is the second
            passed PPM.Study enum
            :param this: The study object to be compared
            :type this: object
            :param that: What we are comparing against
            :type that: object
            :return: Whether they are one and the same
            :rtype: boolean
            """
            # Compare
            return PPM.Study.get(this) is PPM.Study.get(that)

        @staticmethod
        def fhir_id(study):
            """
            Return the FHIR identifier for the passed study
            :return: A PPM study identifier
            :rtype: str
            """
            return "ppm-{}".format(PPM.Study.get(study).value)

        @classmethod
        def identifiers(cls):
            """
            Return a list of all PPM study identifiers to be used in FHIR resources
            :return: A list of PPM study identifiers
            :rtype: list
            """
            return [PPM.Study.fhir_id(study) for study in cls]

        @staticmethod
        def testing(study):
            """
            Return true if the passed study is testing
            :rtype: boolean
            """
            return PPM.Study.get(study) in [PPM.Study.EXAMPLE]

        @staticmethod
        def is_ppm(identifier):
            """
            Returns whether a study identifier is a PPM one or not
            :param identifier: The study identifier
            :type identifier: str
            :return: Whether it's a PPM study or not
            :rtype: bool
            """
            return identifier.lower() in PPM.Study.identifiers()

        @classmethod
        def enum(cls, enum):
            """Accepts any form of an enum and returns the enum"""
            for item in cls:
                if (
                    enum is item
                    or enum == item.name
                    or enum == item.value
                    or enum == cls.title(item)
                    or enum == cls.fhir_id(item)
                ):
                    return item

            raise ValueError('Value "{}" is not a valid {}'.format(enum, cls.__name__))

        @classmethod
        def get(cls, enum):
            """
            Returns an instance of the Study enum for the given study value, enum,
            whatever
            :param enum: The study value string, name or enum
            :type enum: Object
            :return: The instance of PPM Study enum
            :rtype: PPM.Study
            """
            return cls.enum(enum)

        @staticmethod
        def dashboard_url(study):
            """
            Returns the defined dashboard for the passed study
            :param study: The study we want the dashboard URL for
            :type study: PPM.Study
            :return: The dashboard URL
            :rtype: str
            """
            return ppm_settings.dashboard_url(PPM.Study.get(study).value)

        @staticmethod
        def from_value(study):
            """
            Returns an instance of the Study enum for the given study value
            :param study: The study value string
            :type study: str
            :return: The instance of PPM Study enum
            :rtype: PPM.Study
            """
            return PPM.Study.get(study)

        @classmethod
        def title(cls, study):
            """
            Returns the title to be used for the given study.
            :param study: The study identifier or value
            :type study: object
            :return: The title for the study
            :rtype: str
            """
            return dict(PPM.Study.choices())[PPM.Study.get(study).value]

        @classmethod
        def choices(cls):
            """
            Returns a choices tuple of tuples. Define enum titles if different
            from names/values here as this is the source for pulling enum titles.
            :return: ((str, str), )
            """
            return (
                (PPM.Study.NEER.value, "NEER"),
                (PPM.Study.ASD.value, "ASD"),
                (PPM.Study.EXAMPLE.value, "Example"),
                (PPM.Study.RANT.value, "RANT"),
            )

        @classmethod
        def dashboard_url_for_study(cls, study):
            """
            Accepts a valid PPM study and returns the URL for that study's dashboard
            :param study: The study
            :return: str
            """
            url = furl(settings.PPM_DASHBOARD_URL)
            url.path.segments.extend(["dashboard", cls.get(study).value])

            return url.url

        @classmethod
        def dashboard(cls, study, environment):
            """
            This method returns the dashboard step description for the given study
            :param study: The study for which the dashboard will be generated
            :param environment: The environment in which PPM is running
            :return: dict
            """
            # Get the enum
            _study = PPM.Study.get(study)

            steps = None
            if _study is PPM.Study.ASD:
                steps = [
                    {
                        "step": "email-confirm",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "registration",
                        "blocking": True,
                        "required": True,
                        "post_enrollment": PPM.Enrollment.Registered.value,
                        "enabled": True,
                    },
                    {
                        "step": "consent",
                        "blocking": True,
                        "required": True,
                        "pre_enrollment": PPM.Enrollment.Registered.value,
                        "post_enrollment": PPM.Enrollment.Consented.value,
                        "enabled": True,
                    },
                    {
                        "step": "poc",
                        "blocking": True,
                        "required": True,
                        "pre_enrollment": PPM.Enrollment.Consented.value,
                        "post_enrollment": PPM.Enrollment.Proposed.value,
                        "enabled": True,
                    },
                    {
                        "step": "approval",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "questionnaire",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "twitter",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "fitbit",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "facebook",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                ]
            elif _study is PPM.Study.EXAMPLE:
                steps = [
                    {
                        "step": "email-confirm",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "registration",
                        "blocking": True,
                        "required": True,
                        "post_enrollment": PPM.Enrollment.Registered.value,
                        "enabled": True,
                    },
                    {
                        "step": "consent",
                        "blocking": True,
                        "required": True,
                        "pre_enrollment": PPM.Enrollment.Registered.value,
                        "post_enrollment": PPM.Enrollment.Consented.value,
                    },
                    {
                        "step": "questionnaire",
                        "blocking": True,
                        "required": True,
                        "pre_enrollment": PPM.Enrollment.Consented.value,
                        "post_enrollment": PPM.Enrollment.Proposed.value,
                        "enabled": True,
                    },
                    {
                        "step": "approval",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "poc",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "research-studies",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "survey",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "twitter",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "fitbit",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "facebook",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "procure",
                        "blocking": False,
                        "required": False,
                        "multiple": False,
                        "enabled": True,
                    },
                    {
                        "step": "ehr",
                        "blocking": False,
                        "required": False,
                        "multiple": True,
                        "enabled": True,
                    },
                    {
                        "step": "picnichealth",
                        "blocking": False,
                        "required": True,
                        "enabled": True,
                    },
                ]
            elif _study is PPM.Study.NEER:
                steps = [
                    {
                        "step": "email-confirm",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "registration",
                        "blocking": True,
                        "required": True,
                        "post_enrollment": PPM.Enrollment.Registered.value,
                        "enabled": True,
                    },
                    {
                        "step": "consent",
                        "blocking": True,
                        "required": True,
                        "pre_enrollment": PPM.Enrollment.Registered.value,
                        "post_enrollment": PPM.Enrollment.Consented.value,
                        "enabled": True,
                    },
                    {
                        "step": "questionnaire",
                        "blocking": True,
                        "required": True,
                        "pre_enrollment": PPM.Enrollment.Consented.value,
                        "post_enrollment": PPM.Enrollment.Proposed.value,
                        "enabled": True,
                    },
                    {
                        "step": "approval",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "poc",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "research-studies",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "survey",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "picnichealth",
                        "blocking": False,
                        "required": True,
                        "enabled": True,
                    },
                ]
            elif _study is PPM.Study.RANT:
                steps = [
                    {
                        "step": "email-confirm",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "registration",
                        "blocking": True,
                        "required": True,
                        "post_enrollment": PPM.Enrollment.Registered.value,
                        "enabled": True,
                    },
                    {
                        "step": "consent",
                        "blocking": True,
                        "required": True,
                        "pre_enrollment": PPM.Enrollment.Registered.value,
                        "post_enrollment": PPM.Enrollment.Consented.value,
                        "enabled": True,
                    },
                    {
                        "step": "questionnaire",
                        "blocking": True,
                        "required": True,
                        "pre_enrollment": PPM.Enrollment.Consented.value,
                        "post_enrollment": PPM.Enrollment.Proposed.value,
                        "enabled": True,
                    },
                    {
                        "step": "approval",
                        "blocking": True,
                        "required": True,
                        "enabled": True,
                    },
                    {
                        "step": "qualtrics-SV_6YHOgpujEr8uDaZ",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "qualtrics-SV_74G8i7ukzojD88J",
                        "blocking": False,
                        "required": False,
                        "enabled": True,
                    },
                    {
                        "step": "procure",
                        "blocking": False,
                        "required": False,
                        "multiple": False,
                        "enabled": True,
                    },
                    {
                        "step": "picnichealth",
                        "blocking": False,
                        "required": True,
                        "enabled": True,
                    },
                ]

            return steps

        @staticmethod
        def is_dashboard_step_enabled(step, study, environment):
            """
            Returns whether the given step for the current study and environment
            should be enabled or not.
            :param step: The code for the step to check
            :param study: The current PPM study
            :param environment: The current PPM environment
            :return: bool
            """
            step_dict = next((s for s in PPM.Study.dashboard(study, environment) if s["step"] == step.lower()), None)

            # Check if enabled
            if step_dict and step_dict.get("enabled"):
                return True
            elif not step_dict:
                logger.warning('Step "{}" does not exist in dashboard spec for study "{}"'.format(step, study))

            return False

        @staticmethod
        def providers(study, environment):
            """
            Returns a list of enabled providers for the given study and environment
            """
            # Get the enum
            _study = PPM.Study.get(study)

            providers = None
            if _study is PPM.Study.ASD:
                providers = [
                    PPM.Provider.PPM,
                    PPM.Provider.Fitbit,
                    PPM.Provider.Twitter,
                    PPM.Provider.Facebook,
                    PPM.Provider.PicnicHealth,
                    PPM.Provider.Broad,
                    PPM.Provider.File,
                    PPM.Provider.Qualtrics,
                    PPM.Provider.ImmunoSEQ,
                ]
            elif _study is PPM.Study.NEER:
                providers = [
                    PPM.Provider.PPM,
                    PPM.Provider.Twitter,
                    PPM.Provider.PicnicHealth,
                    PPM.Provider.Broad,
                    PPM.Provider.File,
                    PPM.Provider.Qualtrics,
                    PPM.Provider.ImmunoSEQ,
                ]
            elif _study is PPM.Study.RANT:
                providers = [
                    PPM.Provider.PPM,
                    PPM.Provider.PicnicHealth,
                    PPM.Provider.Broad,
                    PPM.Provider.File,
                    PPM.Provider.Qualtrics,
                    PPM.Provider.ImmunoSEQ,
                    PPM.Provider.Procure,
                ]
            elif _study is PPM.Study.EXAMPLE:
                providers = [
                    PPM.Provider.PPM,
                    PPM.Provider.Fitbit,
                    PPM.Provider.Twitter,
                    PPM.Provider.Facebook,
                    PPM.Provider.PicnicHealth,
                    PPM.Provider.Broad,
                    PPM.Provider.File,
                    PPM.Provider.SMART,
                    PPM.Provider.Qualtrics,
                    PPM.Provider.ImmunoSEQ,
                    PPM.Provider.Procure,
                ]

            return providers

        @staticmethod
        def is_provider_enabled(provider, study, environment):
            """
            Returns whether the given provider for the current study and environment
            should be enabled or not.
            :param provider: The code for the provider to check
            :param study: The current PPM study
            :param environment: The current PPM environment
            :return: bool
            """
            _provider = PPM.Provider.get(provider)
            return _provider in PPM.Study.providers(study, environment)

    # Alias Project as Study until we migrate all usages to Study
    Project = Study

    # Set the appropriate participant statuses
    @total_ordering
    @django_enum
    class Enrollment(PPMEnum):
        Registered = "registered"
        Consented = "consented"
        Proposed = "proposed"
        Accepted = "accepted"
        Completed = "completed"
        Pending = "pending"
        Ineligible = "ineligible"
        Terminated = "terminated"

        def __lt__(self, other):

            # Check equality
            if self is other:
                return False

            # Compare states of enrollment
            if self is PPM.Enrollment.Registered:
                return True
            elif self is PPM.Enrollment.Consented:
                return other not in [PPM.Enrollment.Registered]
            elif self is PPM.Enrollment.Proposed:
                return other not in [
                    PPM.Enrollment.Registered,
                    PPM.Enrollment.Consented,
                ]
            elif self in [PPM.Enrollment.Pending, PPM.Enrollment.Ineligible]:
                return other in [
                    PPM.Enrollment.Accepted,
                    PPM.Enrollment.Completed,
                    PPM.Enrollment.Terminated,
                ]
            elif self is PPM.Enrollment.Accepted:
                return other in [PPM.Enrollment.Completed, PPM.Enrollment.Terminated]
            elif self in [PPM.Enrollment.Completed, PPM.Enrollment.Terminated]:
                return False

        @classmethod
        def enum(cls, enum):
            """Accepts any form of an enum and returns the enum"""
            for item in cls:
                if enum is item or enum == item.name or enum == item.value or enum == cls.title(item):
                    return item

            raise ValueError('Value "{}" is not a valid {}'.format(enum, cls.__name__))

        @classmethod
        def get(cls, enum):
            """Accepts any form of an enrollment and returns the enum"""
            return cls.enum(enum)

        @classmethod
        def choices(cls):
            return (
                (PPM.Enrollment.Registered.value, "Registered"),
                (PPM.Enrollment.Consented.value, "Consented"),
                (PPM.Enrollment.Proposed.value, "Proposed"),
                (PPM.Enrollment.Pending.value, "Pending"),
                (PPM.Enrollment.Accepted.value, "Accepted"),
                (PPM.Enrollment.Ineligible.value, "Queue"),
                (PPM.Enrollment.Completed.value, "Completed"),
                (PPM.Enrollment.Terminated.value, "Terminated"),
            )

        @classmethod
        def active_choices(cls):
            return (choice for choice in PPM.Enrollment.choices() if PPM.Enrollment.is_active(choice[0]))

        @classmethod
        def title(cls, enrollment):
            """Returns the value to be used as the enrollment's title"""
            return dict(PPM.Enrollment.choices())[PPM.Enrollment.get(enrollment).value]

        @staticmethod
        def is_active(enrollment):
            """
            Accepts an enrollment and returns whether it is considered 'active'
            or not. This will determine how this flag is set on participant
            records. The state of being active generally means this participant
            is in the enrollment procedure and is still considered for
            acceptance, or if already accepted, should receive notifications,
            requests, and any other communications concerning the study.
            :param enrollment: The enrollment to consider
            :return: boolean
            """
            enrollment = PPM.Enrollment.get(enrollment)

            # Set these here
            return enrollment not in [
                PPM.Enrollment.Ineligible,
                PPM.Enrollment.Terminated,
                PPM.Enrollment.Completed,
            ]

        @classmethod
        def notification_for_enrollment(cls, enrollment):
            """Returns the identifier of a communication to be sent out when
            this enrollment is set"""
            # Branch
            enrollment = PPM.Enrollment.enum(enrollment)
            if enrollment is PPM.Enrollment.Pending:
                return PPM.Communication.ParticipantPending
            elif enrollment is PPM.Enrollment.Ineligible:
                return PPM.Communication.ParticipantIneligible
            elif enrollment is PPM.Enrollment.Accepted:
                return PPM.Communication.ParticipantAccepted

            return None

    class Communication(PPMEnum):
        ParticipantProposed = "participant-proposed"
        ParticipantPending = "participant-pending"
        ParticipantIneligible = "participant-ineligible"
        ParticipantAccepted = "participant-accepted"
        PicnicHealthRegistration = "picnichealth-registration"

        @classmethod
        def enum(cls, enum):
            """Accepts any form of an enum and returns the enum"""
            for item in cls:
                if enum is item or enum == item.name or enum == item.value or enum == cls.title(item):
                    return item

            raise ValueError('Value "{}" is not a valid {}'.format(enum, cls.__name__))

        @classmethod
        def get(cls, enum):
            """Accepts any form of an communication and returns the enum"""
            return cls.enum(enum)

        @classmethod
        def choices(cls):
            return (
                (PPM.Communication.ParticipantProposed.value, "Participant Proposed"),
                (PPM.Communication.ParticipantPending.value, "Participant Pending"),
                (PPM.Communication.ParticipantIneligible.value, "Participant Queued"),
                (PPM.Communication.ParticipantAccepted.value, "Participant Accepted"),
                (
                    PPM.Communication.PicnicHealthRegistration.value,
                    "PicnicHealth Registration",
                ),
            )

        @classmethod
        def title(cls, communication):
            """Returns the value to be used as the communication's title"""
            return dict(PPM.Communication.choices())[PPM.Communication.get(communication).value]

    class Questionnaire(PPMEnum):
        """
        This class manages and describes the various questionnaires and surveys
        participants are expected to take during the course of the study

        NOTE: When adding a new questionnaire, ensure it is set as a class
        variable and the title is also set in `choices()`
        """

        # Pre-approval survey/Questionnaires
        EXAMPLEQuestionnaire = "ppm-example-registration-questionnaire"
        NEERQuestionnaire = "ppm-neer-registration-questionnaire"
        RANTQuestionnaire = "ppm-rant-registration-questionnaire"

        # Post-approval surveys/questionnaires
        ASDQuestionnaire = "ppm-asd-questionnaire"
        RANTPointsOfCareQuestionnaire = "ppm-rant-points-of-care-questionnaire"
        RANTRASymptomQuestionnaire = "ppm-rant-ra-symptom-questionnaire"
        RANTCOVIDQuestionnaire = "ppm-rant-covid-questionnaire"

        # Consents (Current)
        EXAMPLEConsent = "example-signature"
        NEERConsent = "neer-signature-v2"
        ASDGuardianConsentQuestionnaire = "ppm-asd-consent-guardian-quiz"
        ASDIndividualConsentQuestionnaire = "ppm-asd-consent-individual-quiz"
        ASDConsentIndividualSignatureQuestionnaire = "individual-signature-part-1"
        ASDConsentGuardianSignature1Questionnaire = "guardian-signature-part-1"
        ASDConsentGuardianSignature2Questionnaire = "guardian-signature-part-2"
        ASDConsentGuardianSignature3Questionnaire = "guardian-signature-part-3"

        # Older versions
        NEERConsentV1 = "neer-signature"

        @classmethod
        def choices(cls):
            return (
                (PPM.Questionnaire.EXAMPLEQuestionnaire.value, "Example Registration Questionnaire"),
                (PPM.Questionnaire.NEERQuestionnaire.value, "NEER Registration Questionnaire"),
                (PPM.Questionnaire.ASDQuestionnaire.value, "ASD Registration Questionnaire"),
                (PPM.Questionnaire.RANTQuestionnaire.value, "RANT Registration Questionnaire"),
                (PPM.Questionnaire.RANTPointsOfCareQuestionnaire.value, "RANT Points of Care Questionnaire"),
                (PPM.Questionnaire.RANTRASymptomQuestionnaire.value, "RANT RA Symptom Questionnaire"),
                (PPM.Questionnaire.RANTCOVIDQuestionnaire.value, "RANT COVID Questionnaire"),
                (PPM.Questionnaire.EXAMPLEConsent.value, "Example Consent Signature"),
                (PPM.Questionnaire.NEERConsent.value, "NEER Consent Signature"),
                (PPM.Questionnaire.ASDGuardianConsentQuestionnaire.value, "ASD Guardian Consent Questionnaire"),
                (PPM.Questionnaire.ASDIndividualConsentQuestionnaire.value, "ASD Individual ConsentQuestionnaire"),
                (
                    PPM.Questionnaire.ASDConsentIndividualSignatureQuestionnaire.value,
                    "ASD Individual Consent Signature",
                ),
                (PPM.Questionnaire.ASDConsentGuardianSignature1Questionnaire.value, "ASD Guardian Consent Signature"),
                (PPM.Questionnaire.ASDConsentGuardianSignature2Questionnaire.value, "ASD Guardian Consent Signature"),
                (PPM.Questionnaire.ASDConsentGuardianSignature3Questionnaire.value, "ASD Guardian Assent Signature"),
            )

        @classmethod
        def questionnaires_for_study(cls, study):
            """
            Returns a list of Questionnaires for the passed study. This includes
            both mandatory and optional questionnaires.

            :param study: The study to check
            :type study: PPM.Study
            :raises ValueError: Throws error if the study is invalid
            :return: A list of PPM.Questionnaire
            :rtype: list
            """
            # Match questionnaires by ID
            reg = re.compile(r"^ppm-{}[-]?.*-questionnaire$".format(PPM.Study.get(study).value))
            return [q for q in PPM.Questionnaire if type(q.value) is str and reg.fullmatch(q.value) is not None]

        @classmethod
        def extra_questionnaires_for_study(cls, study):
            """
            Returns a list of Questionnaires for the passed study. This includes
            only optional questionnaires that are given post-approval.

            :param study: The study to check
            :type study: PPM.Study
            :raises ValueError: Throws error if the study is invalid
            :return: A list of PPM.Questionnaire
            :rtype: list
            """
            # Match questionnaires by ID
            return list(
                set(PPM.Questionnaire.questionnaires_for_study(study))
                - set(PPM.Questionnaire.questionnaire_for_study(study))
            )

        @classmethod
        def qualtrics_survey_ids(cls):
            """
            This method returns a mapping of Qualtrics survey IDs to PPM
            Questionnaire IDs.

            :return: A dictionary of Qualtrics IDs to PPM Questionnaire IDs
            :rtype: dict
            """
            # Get mapping from environment
            try:
                # Attempt to parse dictionary from environment
                surveys = json.loads(os.environ.get("SOURCE_QUALTRICS_SURVEYS"))

                # Check format
                for study, study_surveys in surveys.items():
                    for survey in study_surveys:
                        try:
                            # Ensure valid questionnaire IDs
                            cls.get(survey["questionnaire_id"])

                        except ValueError as e:
                            logger.exception(
                                "PPM/Questionnaire: Qualtrics survey mapping is"
                                " invalid, unknown key: {} : {}".format(survey, e),
                                exc_info=True,
                            )
                return surveys

            except Exception as e:
                logger.exception("PPM/Questionnaire: Error parsing Qualtrics " "map: {}".format(e), exc_info=True)

            return {}

        @classmethod
        def questionnaire_id_for_qualtrics_id(cls, survey_id):
            """
            This method links Qualtrics surveys to PPM questionnaire resources.
            This is only used for surveys that are administered by Qualtrics
            but are sent to PPM for storage (typically in FHIR) to be used
            elsewhere in the dashboards.

            :param survey_id: The Qualtrics survey ID
            :type survey_id: str
            :raises ValueError: If either ID is invalid, ValueError is raied
            :return: The ID of the related PPM questionnaire resource in FHIR
            :rtype: str
            """
            try:
                for _, surveys in cls.qualtrics_survey_ids().items():

                    # Check for questionnaire ID
                    questionnaire_id = next(
                        (s.get("questionnaire_id") for s in surveys if s.get("survey_id") == survey_id), None
                    )

                    # If not questionnaire ID
                    if not questionnaire_id:
                        continue

                    # Return it
                    logger.debug('PPM/Questionnaire: Found "{}"' 'for "{}"'.format(questionnaire_id, survey_id))

                    return cls.get(questionnaire_id).value

            except Exception as e:
                logger.exception("Error: Unknown Qualtrics survey " '"{}": {}'.format(survey_id, e), exc_info=True)

            return None

        @classmethod
        def qualtrics_id_for_questionnaire_id(cls, questionnaire_id):
            """
            This method links Qualtrics surveys to PPM questionnaire resources.
            This is only used for surveys that are administered by Qualtrics
            but are sent to PPM for storage (typically in FHIR) to be used
            elsewhere in the dashboards.

            :param questionnaire_id: The PPM questionnaire ID
            :type questionnaire_id: str
            :raises ValueError: If either ID is invalid, ValueError is raied
            :return: The ID of the related Qualtrics Survey
            :rtype: str
            """
            try:
                for _, surveys in cls.qualtrics_survey_ids().items():

                    # Check for questionnaire ID
                    survey_id = next(
                        (s.get("survey_id") for s in surveys if s.get("questionnaire_id") == questionnaire_id), None
                    )

                    # If not survey ID
                    if not survey_id:
                        continue

                    # Return it
                    logger.debug('PPM/Questionnaire: Found "{}" for "{}"'.format(survey_id, questionnaire_id))

                    return survey_id

            except Exception as e:
                logger.exception('Error: Unknown PPM Questionnaire "{}": {}'.format(questionnaire_id, e), exc_info=True)

            return None

        @staticmethod
        def questionnaire_url_for_study(study, return_url=None):
            """
            Returns the URL of the Questionnaire to be taken by participants
            during registration.
            :param study: The PPM study
            :return: str
            """
            url = furl(os.environ["PPM_QUESTIONNAIRE_URL"])

            # Add components
            url.path.set("fhirquestionnaire/questionnaire/p/{}/".format(study.value))

            # Check for return URL
            if return_url:
                url.query.params.add("return_url", return_url)

            return url.url

        @classmethod
        def questionnaire_url_for_questionnaire(cls, questionnaire, return_url=None):
            """
            Returns the URL of the Questionnaire to be taken by participants
            typically after approval and during general data collection.
            :param study: The PPM study
            :return: str
            """
            # Ensure we're working with the enum
            questionnaire = cls.get(questionnaire)

            # Get the base URL for the service
            url = furl(os.environ["PPM_QUESTIONNAIRE_URL"])

            # Get study this questionnaire pertains to
            study = next(s for s in PPM.Study if questionnaire in cls.questionnaires_for_study(s))

            # Add components
            url.path.set("fhirquestionnaire/questionnaire/q/{}/{}/".format(study.value, questionnaire.value))

            # Check for return URL
            if return_url:
                url.query.params.add("return_url", return_url)

            return url.url

        @staticmethod
        def consent_url_for_study(study):
            """
            Returns the URL of the consent to be taken by participants during
            registration.
            :param study: The PPM study
            :return: str
            """
            url = furl(os.environ["PPM_QUESTIONNAIRE_URL"])

            # Add components
            url.path.set("fhirquestionnaire/consent/p/{}/".format(study.value))

            return url.url

        @classmethod
        def consent_questionnaire_for_study(cls, study, **kwargs):
            """
            Returns the FHIR ID of the Questionnaire resource that is used as the
            signature to be filled out by participants during the consent step
            of the sign up process.
            :param study: The PPM study
            :return: str
            """
            if PPM.Study.get(study) is PPM.Study.ASD:

                # We need more info
                if kwargs.get("type") == "guardian":
                    return (
                        cls.ASDConsentGuardianSignature1Questionnaire.value,
                        cls.ASDConsentGuardianSignature2Questionnaire.value,
                        cls.ASDConsentGuardianSignature3Questionnaire.value,
                    )
                else:
                    return PPM.Questionnaire.ASDIndividualConsentQuestionnaire.value

            elif PPM.Study.get(study) is PPM.Study.NEER:
                return PPM.Questionnaire.NEERConsent.value

            elif PPM.Study.get(study) is PPM.Study.RANT:
                return None

            elif PPM.Study.get(study) is PPM.Study.EXAMPLE:
                return PPM.Questionnaire.EXAMPLEConsent.value

        @staticmethod
        def questionnaire_for_study(study):
            """
            Returns the FHIR ID of the Questionnaire resource that is used as the
            survey to be filled out by participants during sign up.
            :param study: The PPM study
            :return: str
            """
            if PPM.Study.get(study) is PPM.Study.ASD:
                return PPM.Questionnaire.ASDQuestionnaire.value

            elif PPM.Study.get(study) is PPM.Study.NEER:
                return PPM.Questionnaire.NEERQuestionnaire.value

            elif PPM.Study.get(study) is PPM.Study.RANT:
                return PPM.Questionnaire.RANTQuestionnaire.value

            elif PPM.Study.get(study) is PPM.Study.EXAMPLE:
                return PPM.Questionnaire.EXAMPLEQuestionnaire.value

        @staticmethod
        def questionnaire_for_consent(composition):
            """
            Returns the FHIR ID of the Questionnaire resource that is used as the
            quiz to be filled out by participants during the consent process.
            :param composition: The consent composition resource
            :return: str
            """
            if composition.get("type", "").lower() == "guardian":
                return PPM.Questionnaire.ASDGuardianConsentQuestionnaire.value

            else:
                return PPM.Questionnaire.ASDIndividualConsentQuestionnaire.value

        @staticmethod
        def exceptions(questionnaire_id):
            """
            Returns a dictionary mapping the question link ID to the SNOMED code
            to use for the exclusion item
            :param questionnaire_id: The Questionnaire ID
            :return: dict
            """
            if questionnaire_id == PPM.Questionnaire.NEERConsent.value:
                return {
                    "question-1": "82078001",
                    "question-2": "258435002",
                    "question-3": "284036006",
                    "question-4": "702475000",
                }

            elif questionnaire_id == PPM.Questionnaire.RANTConsent.value:
                return {}

            elif questionnaire_id == PPM.Questionnaire.EXAMPLEConsent.value:
                return {
                    "question-1": "82078001",
                    "question-2": "165334004",
                    "question-3": "258435002",
                    "question-4": "284036006",
                    "question-5": "702475000",
                }

            elif questionnaire_id == PPM.Questionnaire.ASDConsentIndividualSignatureQuestionnaire.value:
                return {
                    "question-1": "225098009",
                    "question-2": "284036006",
                    "question-3": "702475000",
                }

            elif questionnaire_id == PPM.Questionnaire.ASDGuardianConsentQuestionnaire.value:
                return {
                    "question-1": "225098009",
                    "question-2": "284036006",
                    "question-3": "702475000",
                }

            elif questionnaire_id == PPM.Questionnaire.ASDConsentIndividualSignatureQuestionnaire.value:
                return {
                    "question-1": "225098009",
                    "question-2": "284036006",
                }

            else:
                raise ValueError(
                    'Questionnaire ID "{}" is either not a valid PPM '
                    "consent Questionnaire, or its exception mappings has not "
                    "yet been added.".format(questionnaire_id)
                )

        @staticmethod
        def questionnaire_for_project(project):  # TODO: Deprecated, remove!
            return PPM.Questionnaire.questionnaire_for_study(project)

    class Provider(Enum):
        PPM = "ppmfhir"
        Fitbit = "fitbit"
        Twitter = "twitter"
        Facebook = "facebook"
        Gencove = "gencove"
        uBiome = "ubiome"
        PicnicHealth = "picnichealth"
        Broad = "broad"
        SMART = "smart"
        File = "file"
        Qualtrics = "qualtrics"
        ImmunoSEQ = "immunoseq"
        Procure = "procure"

        @classmethod
        def enum(cls, enum):
            """Accepts any form of an enum and returns the enum"""
            for item in cls:
                if enum is item or enum == item.name or enum == item.value or enum == cls.title(item):
                    return item

            raise ValueError('Value "{}" is not a valid {}'.format(enum, cls.__name__))

        @classmethod
        def get(cls, enum):
            """Accepts any form of an provider and returns the enum"""
            return cls.enum(enum)

        @classmethod
        def choices(cls):
            return (
                (PPM.Provider.PPM.value, "PPM FHIR"),
                (PPM.Provider.Fitbit.value, "Fitbit"),
                (PPM.Provider.Twitter.value, "Twitter"),
                (PPM.Provider.Facebook.value, "Facebook"),
                (PPM.Provider.Gencove.value, "Gencove"),
                (PPM.Provider.uBiome.value, "uBiome"),
                (PPM.Provider.Broad.value, "Broad"),
                (PPM.Provider.PicnicHealth.value, "PicnicHealth"),
                (PPM.Provider.SMART.value, "SMART on FHIR"),
                (PPM.Provider.File.value, "PPM Files"),
                (PPM.Provider.Qualtrics.value, "Qualtrics Surveys"),
                (PPM.Provider.ImmunoSEQ.value, "immunoSEQ"),
                (PPM.Provider.Procure.value, "Procure"),
            )

        @classmethod
        def title(cls, provider):
            """Returns the value to be used as the provider's title"""
            return dict(PPM.Provider.choices())[PPM.Provider.get(provider).value]

    class TrackedItem(Enum):
        Fitbit = "fitbit"
        SalivaSampleKit = "spitkit"
        uBiomeFecalSampleKit = "ubiome"
        BloodSampleKit = "blood"

        @classmethod
        def enum(cls, enum):
            """Accepts any form of an enum and returns the enum"""
            for item in cls:
                if enum is item or enum == item.name or enum == item.value or enum == cls.title(item):
                    return item

            raise ValueError('Value "{}" is not a valid {}'.format(enum, cls.__name__))

        @classmethod
        def get(cls, enum):
            """Accepts any form of an tracked item and returns the enum"""
            return cls.enum(enum)

        @classmethod
        def choices(cls):
            return (
                (PPM.TrackedItem.Fitbit.value, "FitBit"),
                (PPM.TrackedItem.SalivaSampleKit.value, "Saliva Kit"),
                (PPM.TrackedItem.uBiomeFecalSampleKit.value, "uBiome"),
                (PPM.TrackedItem.BloodSampleKit.value, "Blood Sample"),
            )

        @classmethod
        def title(cls, tracked_item):
            """
            Returns the title for the given tracked item/device
            :param tracked_item: The item code/ID
            :type tracked_item: str
            :return: The item's title
            :rtype: str
            """
            return dict(PPM.TrackedItem.choices())[PPM.TrackedItem.get(tracked_item).value]

        @staticmethod
        def devices(study=None):
            """
            Returns the device item codes for every project in PPM
            :param study: The study for which the devices should be returned
            :return: A list of device codes
            :rtype: list
            """
            devices = {
                PPM.Study.NEER.value: [
                    PPM.TrackedItem.Fitbit.value,
                    PPM.TrackedItem.uBiomeFecalSampleKit.value,
                    PPM.TrackedItem.BloodSampleKit.value,
                ],
                PPM.Study.RANT.value: [
                    PPM.TrackedItem.BloodSampleKit.value,
                ],
                PPM.Study.ASD.value: [
                    PPM.TrackedItem.Fitbit.value,
                    PPM.TrackedItem.SalivaSampleKit.value,
                ],
                PPM.Study.EXAMPLE.value: [
                    PPM.TrackedItem.Fitbit.value,
                    PPM.TrackedItem.uBiomeFecalSampleKit.value,
                    PPM.TrackedItem.BloodSampleKit.value,
                    PPM.TrackedItem.SalivaSampleKit.value,
                ],
            }

            return devices[study] if study else devices

    # Alias Device to TrackedItem
    Device = TrackedItem

    class Email(Enum):
        AdminProposedNotification = "admin_proposed_notification"
        AdminContactNotification = "admin_contact_notification"
        UserProposedNotification = "user_proposed_notification"
        UserAcceptedNotification = "user_accepted_notification"
        UserPendingNotification = "user_pending_notification"
        UserQueuedNotification = "user_queued_notification"

        @staticmethod
        def send(email, study, recipients, subject=None, sender=None, context=None, reply_to=None):
            """
            Sends an email for the corresponding email identifier.
            :param email: The email identifier
            :param study: The study for which the email concerns
            :param recipients: A list of recipients
            :param subject: The subject for the email
            :param sender: The sender
            :param context: Context for the email
            :param reply_to: A list of reply to addresses, if any
            :return:
            """
            try:
                # Add subject
                if not subject:
                    subject = PPM.Email.subject(email=email, study=study)

                for recipient in recipients:
                    logger.debug(f'Email: Sending "{email.value}" for study "{study}" -> "{recipient}"')

                    # Check if test
                    test_admin = PPM.Email.is_test(recipient)
                    if test_admin:
                        recipient = test_admin

                    # Check reply to
                    if not reply_to:
                        reply_to = []

                    # Render templates
                    html, plain = PPM.Email.render(email=email, study=study, context=context, subject=subject)

                    # Perform send
                    msg = EmailMultiAlternatives(
                        subject, plain, ppm_settings.EMAIL_DEFAULT_FROM, [recipient], reply_to=reply_to
                    )
                    msg.attach_alternative(html, "text/html")
                    msg.send()

                    logger.debug(f'Email: Sent email "{email.value}" for study "{study}" -> "{recipient}"')

                    return True

            except Exception as e:
                logger.exception(
                    "Email error: {}".format(e),
                    exc_info=True,
                    extra={
                        "email": email.value,
                        "study": study,
                        "subject": subject,
                        "sender": sender,
                        "context": context,
                    },
                )

            return False

        @staticmethod
        def is_test(recipient):
            """
            Tests the given user account for matching that of a test account.
            Test accounts are specified as [regex 1]:[admin email],[regex 2]:[admin email],...
            If the given user is a test account, return the specific admin for which notifications
            should be limited to sending to.
            :param recipient: The recipient email
            :return: Admin email address if test account, None if not
            """

            # Check for test accounts
            try:
                if hasattr(ppm_settings, "EMAIL_TEST_ACCOUNTS"):
                    test_accounts = ppm_settings.EMAIL_TEST_ACCOUNTS
                elif hasattr(settings, "TEST_EMAIL_ACCOUNTS"):
                    test_accounts = settings.TEST_EMAIL_ACCOUNTS.split(",")
                else:
                    return None

                if test_accounts is not None and len(test_accounts) > 0:
                    logger.info("Test accounts found, checking now...")

                    for test_account in test_accounts:

                        # Split the test account email from the destination admin email
                        test_account_parts = test_account.split(":")
                        regex = re.compile(test_account_parts[0])
                        matches = regex.match(recipient)
                        if matches is not None and matches.group():
                            logger.info(
                                "Email: Test account found: {}, sending to {}".format(recipient, test_account_parts[1])
                            )

                            # Return the test admin email
                            return test_account_parts[1]

            except Exception as e:
                logger.warning("Test email lookup failed: {}".format(e), exc_info=True)

            return None

        @staticmethod
        def render(email, study, context, subject=None):
            """
            Accepts an email identifier and context and returns the rendered content as a string
            :param email: The email identifier
            :param study: The study for which the email concerns
            :param context: The context for the email's content
            :param subject: The optional subject line
            :return: str
            """
            # Get the study
            _study = PPM.Study.get(study)

            # Add study
            context["ppm_study"] = _study.value
            context["ppm_study_title"] = PPM.Study.title(study)

            # Add subject
            context["ppm_subject"] = subject if subject is not None else PPM.Email.subject(email=email, study=study)

            # Add signature bits
            context["ppm_signature"] = ppm_settings.EMAIL_SIGNATURE

            try:
                # Check for study specific templates
                if os.path.exists(f"ppmutils/{_study.value}/{email.value}.html"):

                    # These templates are specific to this study
                    template_paths = (
                        f"ppmutils/{_study.value}/{email.value}.html",
                        f"ppmutils/{_study.value}/{email.value}.txt",
                    )

                else:

                    # These are generic templates and can be rendered across all studies
                    template_paths = f"ppmutils/{email.value}.html", f"ppmutils/{email.value}.txt"

                # Render templates
                html = render_to_string(template_paths[0], context)
                plain = render_to_string(template_paths[1], context)

                return html, plain

            except Exception as e:
                logger.exception(
                    f"Email error: {e}",
                    exc_info=True,
                    extra={
                        "email": email.value,
                        "study": study,
                    },
                )

        @staticmethod
        def subject(email, study):
            """
            Returns the default subject time for the email and study combination
            :param email: The email identifier
            :param study: The study for which the email concerned
            :return: str
            """
            # Set email subject lines
            subjects = {
                "admin_contact_notification": f"People-Powered Medicine - {PPM.Study.title(study)} - Support",
                "admin_proposed_notification": f"People-Powered Medicine - {PPM.Study.title(study)} - New User Signup",
                "user_proposed_notification": f"People-Powered Medicine - {PPM.Study.title(study)} - Registration",
                "user_accepted_notification": f"People-Powered Medicine - {PPM.Study.title(study)} - Approved",
                "user_queued_notification": f"People-Powered Medicine - {PPM.Study.title(study)} - Update",
                "user_pending_notification": f"People-Powered Medicine - {PPM.Study.title(study)} - Update",
            }

            return subjects[email.value]

    class Service(object):

        # Subclasses set this to direct requests
        service = None

        # If enabled, all requests are routed through PPM-API
        proxied = False

        # Set some auth header properties
        jwt_cookie_name = "DBMI_JWT"
        jwt_authorization_prefix = "JWT"
        token_authorization_prefix = "Token"
        ppm_settings_url_name = None

        @classmethod
        def _build_url(cls, path):

            # Build the url, chancing on doubling up a slash or two.
            scheme, domain = tuple(cls.service_url().strip("/").split("://"))
            url = furl("{}://{}/{}".format(scheme, domain, path.lstrip("/").replace("//", "/")))

            return url.url

        @classmethod
        def service_url(cls):

            # Check if this service should be proxied through PPM-API
            if cls.proxied:

                # We want only the domain and no paths, as those should be specified in the calls
                # so strip any included paths and queries and return
                url = furl(ppm_settings.API_URL)
                url.path.segments.clear()
                url.query.params.clear()

                # Add API base path
                url.path.segments.extend(["api", "v1", cls.service])

            else:
                # Get from ppm settings
                if not hasattr(ppm_settings, cls.ppm_settings_url_name):
                    raise SystemError("Service URL not defined in settings".format(cls.service.upper()))

                # Get it
                service_url = getattr(ppm_settings, cls.ppm_settings_url_name)

                # We want only the domain and no paths, as those should be specified in the calls
                # so strip any included paths and queries and return
                url = furl(service_url)
                url.path.segments.clear()
                url.query.params.clear()

            return url.url

        @classmethod
        def headers(cls, request=None, content_type="application/json"):
            """
            Builds request headers. If no request is passed, service is assumed to
            use a pre-defined token in settings as `[SERVICE_NAME]_AUTH_TOKEN`
            :param request: The current request, if any
            :param content_type: The request content type, defaults to JSON
            :return: dict
            """
            if request and cls.get_jwt(request):

                # Use JWT
                return {
                    "Authorization": "{} {}".format(cls.jwt_authorization_prefix, cls.get_jwt(request)),
                    "Content-Type": content_type,
                }

            elif hasattr(settings, "{}_AUTH_TOKEN".format(cls.service.upper())):

                # Get token
                token = getattr(settings, "{}_AUTH_TOKEN".format(cls.service.upper()))

                # Check for specified prefix
                prefix = getattr(
                    settings,
                    "{}_AUTH_PREFIX".format(cls.service.upper()),
                    cls.token_authorization_prefix,
                )

                # Use token
                return {
                    "Authorization": "{} {}".format(prefix, token),
                    "Content-Type": content_type,
                }

            raise SystemError(
                'No request with JWT, or no token specified for service "{}", '
                "cannot build request headers".format(cls.service)
            )

        @classmethod
        def get_jwt(cls, request):

            # Get the JWT token depending on request type
            if hasattr(request, "COOKIES") and request.COOKIES.get(cls.jwt_cookie_name):
                return request.COOKIES.get(cls.jwt_cookie_name)

            # Check if JWT in HTTP Authorization header
            elif (
                hasattr(request, "META")
                and request.META.get("HTTP_AUTHORIZATION")
                and cls.jwt_authorization_prefix in request.META.get("HTTP_AUTHORIZATION")
            ):

                # Remove prefix and return the token
                return request.META.get("HTTP_AUTHORIZATION").replace("{} ".format(cls.jwt_authorization_prefix), "")

            return None

        @classmethod
        def head(cls, request=None, path="/", data=None, raw=False):
            """
            Runs the appropriate REST operation. Request is required for JWT auth,
            not required for token auth.
            :param request: The current Django request
            :param path: The path of the request
            :param data: Request data or params
            :param raw: How the response should be returned
            :return: object
            """
            # Check for params
            if not data:
                data = {}

            try:
                # Prepare the request.
                response = requests.head(cls._build_url(path), headers=cls.headers(request), params=data)

                # Check response type
                if raw:
                    return response
                else:
                    return response.json()

            except Exception as e:
                logger.exception(
                    "{} error: {}".format(cls.service, e),
                    exc_info=True,
                    extra={
                        "data": data,
                        "path": path,
                    },
                )

            return None

        @classmethod
        def get(cls, request=None, path="/", data=None, raw=False):
            """
            Runs the appropriate REST operation. Request is required for JWT auth,
            not required for token auth.
            :param request: The current Django request
            :param path: The path of the request
            :param data: Request data or params
            :param raw: How the response should be returned
            :return: object
            """
            # Check for params
            if not data:
                data = {}

            try:
                # Prepare the request.
                response = requests.get(cls._build_url(path), headers=cls.headers(request), params=data)

                # Check response type
                if raw:
                    return response
                else:
                    return response.json()

            except Exception as e:
                logger.exception(
                    "{} error: {}".format(cls.service, e),
                    exc_info=True,
                    extra={
                        "data": data,
                        "path": path,
                    },
                )

            return None

        @classmethod
        def post(cls, request=None, path="/", data=None, raw=False):
            """
            Runs the appropriate REST operation. Request is required for JWT auth,
            not required for token auth.
            :param request: The current Django request
            :param path: The path of the request
            :param data: Request data or params
            :param raw: How the response should be returned
            :return: object
            """
            # Check for params
            if not data:
                data = {}

            try:
                # Prepare the request.
                response = requests.post(
                    cls._build_url(path),
                    headers=cls.headers(request),
                    data=json.dumps(data),
                )

                # Check response type
                if raw:
                    return response
                else:
                    return response.json()

            except Exception as e:
                logger.exception(
                    "{} error: {}".format(cls.service, e),
                    exc_info=True,
                    extra={
                        "data": data,
                        "path": path,
                    },
                )

            return None

        @classmethod
        def put(cls, request=None, path="/", data=None, raw=False):
            """
            Runs the appropriate REST operation. Request is required for JWT auth,
            not required for token auth.
            :param request: The current Django request
            :param path: The path of the request
            :param data: Request data or params
            :param raw: How the response should be returned
            :return: object
            """
            # Check for params
            if not data:
                data = {}

            try:
                # Prepare the request.
                response = requests.put(
                    cls._build_url(path),
                    headers=cls.headers(request),
                    data=json.dumps(data),
                )

                # Check response type
                if raw:
                    return response
                else:
                    return response.json()

            except Exception as e:
                logger.exception(
                    "{} error: {}".format(cls.service, e),
                    exc_info=True,
                    extra={
                        "data": data,
                        "path": path,
                    },
                )

            return None

        @classmethod
        def patch(cls, request=None, path="/", data=None, raw=False):
            """
            Runs the appropriate REST operation. Request is required for JWT auth,
            not required for token auth.
            :param request: The current Django request
            :param path: The path of the request
            :param data: Request data or params
            :param raw: How the response should be returned
            :return: object
            """
            # Check for params
            if not data:
                data = {}

            try:
                # Prepare the request.
                response = requests.patch(
                    cls._build_url(path),
                    headers=cls.headers(request),
                    data=json.dumps(data),
                )

                # Check response type
                if raw:
                    return response
                else:
                    return response.ok

            except Exception as e:
                logger.exception(
                    "{} error: {}".format(cls.service, e),
                    exc_info=True,
                    extra={
                        "data": data,
                        "path": path,
                    },
                )

            return False

        @classmethod
        def delete(cls, request=None, path="/", data=None, raw=False):
            """
            Runs the appropriate REST operation. Request is required for JWT auth,
            not required for token auth.
            :param request: The current Django request
            :param path: The path of the request
            :param data: Request data or params
            :param raw: How the response should be returned
            :return: object
            """
            # Check for params
            if not data:
                data = {}

            try:
                # Prepare the request.
                response = requests.delete(
                    cls._build_url(path),
                    headers=cls.headers(request),
                    data=json.dumps(data),
                )

                # Check response type
                if raw:
                    return response
                else:
                    return response.ok

            except Exception as e:
                logger.exception(
                    "{} error: {}".format(cls.service, e),
                    exc_info=True,
                    extra={
                        "path": path,
                    },
                )

            return False

        @classmethod
        def request(cls, verb, headers=None, request=None, path="/", data=None, check=True):
            """
            Runs the appropriate REST operation. Request is required for JWT auth,
            not required for token auth.
            :param verb: The RESTful operation to be performed
            :param headers: Any additional headers to include
            :param request: The current Django request
            :param path: The path of the request
            :param data: Request data or params
            :param check: Check the response and raise exception if faulty
            :return: object
            """
            # Check for params
            if not data:
                data = {}

            # Compile headers
            if headers:
                cls.headers(request).update(headers)
            else:
                headers = cls.headers(request)

            # Track response for error reporting
            response = None
            try:
                # Build arguments
                args = [cls._build_url(path)]
                kwargs = {"headers": headers}

                # Check how data should be passed
                if verb.lower() in ["get", "head"]:
                    # Pass dict along
                    kwargs["params"] = data
                else:
                    # Format as JSON string
                    kwargs["data"] = json.dumps(data)

                # Prepare the request.
                response = getattr(requests, verb)(*args, **kwargs)

                # See if we should check the response
                if check:
                    response.raise_for_status()

                # Return
                return response

            except Exception as e:
                logger.exception(
                    "{} {} error: {}".format(cls.service, verb.upper(), e),
                    exc_info=True,
                    extra={
                        "path": path,
                        "verb": verb,
                        "data": data,
                        "response": response,
                    },
                )

            return False
